import json
from pathlib import Path
from datetime import datetime, timezone
from typing import NamedTuple, List, Iterator, Tuple, Union, Any

from ..paths import LocalDir
from ..list_type import ListType


class HistoryEntry(NamedTuple):
    at: datetime
    number: int  # episode or chapter number


class History(NamedTuple):
    mal_id: int
    list_type: str
    title: str
    entries: List[HistoryEntry]

    @property
    def url(self) -> str:
        return f"https://myanimelist.net/{self.list_type}/{self.mal_id}"


def iter_user_history(username: str) -> Iterator[History]:
    localdir = LocalDir.from_username(username)
    yield from iter_history_from_dir(localdir.data_dir)


def iter_history_from_dir(data_dir: Path) -> Iterator[History]:
    # i.e. for anime / manga
    for _type in map(str.lower, ListType.__members__):
        merged_history_file = data_dir / f"{_type}_history.json"
        # parse from both merged history and individual history files,
        # incase either one is missing
        yield from _parse_merged_history(merged_history_file, _type)
        yield from parse_history_dir(data_dir / "history" / _type, _type)


def _parse_merged_history(
    merged_history_file: Path, list_type: Union[str, ListType]
) -> Iterator[History]:
    lt: str = list_type.value.lower() if isinstance(list_type, ListType) else list_type
    if not merged_history_file.exists():
        return
    merged_data = json.loads(merged_history_file.read_text())
    for key, data in merged_data.items():
        title, entries = _parse_history_data(data)
        if len(entries) == 0:
            continue
        yield History(
            list_type=lt,
            mal_id=int(key),
            title=title,
            entries=entries,
        )


def parse_history_dir(
    history_dir: Path, list_type: Union[str, ListType]
) -> Iterator[History]:
    if not history_dir.exists():
        return
    lt: str = list_type.value.lower() if isinstance(list_type, ListType) else list_type
    for history_path in history_dir.glob("*.json"):
        assert (
            history_path.stem.isnumeric()
        ), f"Expected history JSON file, found {history_path}"
        data = json.loads(history_path.read_text())
        title, entries = _parse_history_data(data)
        # only return items which have at least one history entry
        if len(entries) == 0:
            continue
        yield History(
            list_type=lt,
            mal_id=int(history_path.stem),
            title=title,
            entries=entries,
        )


def _parse_history_data(history_data: Any) -> Tuple[str, List[HistoryEntry]]:
    entries: List[HistoryEntry] = []
    for entry_data in history_data["episodes"]:
        [num, epoch] = entry_data
        entries.append(
            HistoryEntry(
                at=datetime.fromtimestamp(epoch, tz=timezone.utc),
                number=num,
            )
        )
    return history_data["title"], entries



import sys
import json
from typing import List, Set, NamedTuple, Callable, Tuple, Optional
from pathlib import Path

from git.repo.base import Repo  # type: ignore[import]
from git.cmd import Git  # type: ignore[import]

from .combine import combine, CombineResults, AnimeData, MangaData
from ..paths import mal_id_cache_dir
from ..log import logger

repo_dir = Path(mal_id_cache_dir)


class Approved(NamedTuple):
    """
    Uses https://github.com/seanbreckenridge/mal-id-cache to fetch a list
    of approved MAL IDs
    """

    anime: Set[int]
    manga: Set[int]

    @staticmethod
    def git_clone() -> Repo:
        """
        Clone or return if .git repo already exists
        """
        if not repo_dir.parent.exists():
            repo_dir.parent.mkdir(exist_ok=True, parents=True)

        if repo_dir.exists():
            assert (
                repo_dir / ".git"
            ).exists(), f"{repo_dir} exists but {repo_dir}/.git does not"
            return Repo(mal_id_cache_dir)

        else:
            return Repo.clone_from(
                "https://github.com/seanbreckenridge/mal-id-cache", repo_dir
            )

    @classmethod
    def git_pull(cls) -> str:
        cls.git_clone()
        gg = Git(repo_dir)
        # TODO: add try/except incase theres no internet
        gg.pull()
        commit_id = str(gg.log().splitlines()[0].split()[-1])
        logger.debug(f"mal-id-cache Commit ID: {commit_id}")
        return commit_id

    @classmethod
    def create_if_doesnt_exist(cls) -> None:
        if not (repo_dir / ".git").exists():
            cls.git_pull()

    @classmethod
    def parse_from_git_dir(cls) -> "Approved":
        cls.create_if_doesnt_exist()
        anime_file = repo_dir / "cache" / "anime_cache.json"
        manga_file = repo_dir / "cache" / "manga_cache.json"
        anime = json.loads(anime_file.read_text())
        manga = json.loads(manga_file.read_text())
        return Approved(
            anime=set(anime["sfw"] + anime["nsfw"]),
            manga=set(manga["sfw"] + manga["nsfw"]),
        )


def _default_parse_func(backup_dir: Path, username: str) -> CombineResults:
    return combine(username, data_dir=backup_dir)


def recover_deleted_single(
    *,
    approved: Approved,
    from_backup_dir: Path,
    username: str,
    filter_with_activity: bool = False,
    parse_func: Optional[Callable[[Path, str], CombineResults]] = None,
) -> CombineResults:
    """
    returns any data from a backup that is not in the approved list
    """

    if parse_func is None:
        parse_func = _default_parse_func

    assert from_backup_dir.exists(), f"Backup dir {from_backup_dir} does not exist"

    new_anime, new_manga = parse_func(from_backup_dir, username)

    deleted_anime = [new for new in new_anime if new.id not in approved.anime]
    deleted_manga = [new for new in new_manga if new.id not in approved.manga]

    if filter_with_activity:
        deleted_anime = [a for a in deleted_anime if len(a.history) > 0]
        deleted_manga = [m for m in deleted_manga if len(m.history) > 0]

    return deleted_anime, deleted_manga


def recover_deleted(
    *,
    approved: Approved,
    backups: List[Path],
    username: str,
    filter_with_activity: bool = False,
    parse_func: Optional[Callable[[Path, str], CombineResults]] = None,
) -> CombineResults:
    """
    parses each backup in reverse order using parse_func, and returns the
    most recent data for each deleted entry

    if filter_with_activity is True, this only returns items which
    have some history activity (watched/read episodes/chapters)
    """

    if parse_func is None:
        try:
            import my.core.structure
        except ImportError:
            logger.error(
                f"my.core.structure module not found, cannot parse backups, install by running `{sys.executable} -m pip install hpi`",
            )
            sys.exit(1)
        else:

            def _parse_func(backup_dir: Path, username: str) -> CombineResults:
                with my.core.structure.match_structure(
                    backup_dir,
                    (
                        "forum",
                        "messages",
                        "friends.json",
                        "animelist.json",
                        "mangalist.json",
                        "animelist.xml",
                        "mangalist.xml",
                        "animelist_api.json",
                        "mangalist_api.json",
                    ),
                    partial=True,
                ) as res:
                    assert len(res) == 1, f"Expected 1 match, got {len(res)}: {res}"
                    return combine(username, data_dir=res[0])

            assert _parse_func is not None
            parse_func = _parse_func

    assert parse_func is not None

    emitted_entries: Set[Tuple[int, str]] = set()
    emit_anime: List[AnimeData] = []
    emit_manga: List[MangaData] = []

    for backup in reversed(backups):
        logger.debug(f"Recovering deleted entries from {backup}")
        anime, manga = recover_deleted_single(
            approved=approved,
            from_backup_dir=backup,
            username=username,
            parse_func=parse_func,
            filter_with_activity=filter_with_activity,
        )
        for a in anime:
            if (a.id, "anime") not in emitted_entries:
                emit_anime.append(a)
                emitted_entries.add((a.id, "anime"))
        for m in manga:
            if (m.id, "manga") not in emitted_entries:
                emit_manga.append(m)
                emitted_entries.add((m.id, "manga"))

    return emit_anime, emit_manga



import json
from typing import NamedTuple, List, Optional, TypeVar, Iterator, Dict, Any
from datetime import date, datetime

from .common import parse_date_safe
from ..list_type import ListType
from ..common import Json
from ..paths import PathIsh, _expand_file
from .mal_list import IdInfo, Season

T = TypeVar("T")


class Entry(NamedTuple):
    entry_type: ListType
    id: int
    title: str
    main_picture: Dict[str, str]
    alternative_titles: Dict[str, Any]
    start_date: Optional[date]
    end_date: Optional[date]
    synopsis: str
    mean: Optional[float]
    rank: Optional[int]
    popularity: int
    num_list_users: int
    num_scoring_users: int
    nsfw: str
    created_at: datetime
    updated_at: datetime
    media_type: str
    status: str
    genres: List[IdInfo]
    list_status: Dict[str, Any]
    episode_count: Optional[int]
    season: Optional[Season]
    source: Optional[str]
    average_episode_duration: Optional[int]
    rating: Optional[str]
    studios: List[IdInfo]

    @classmethod
    def _parse(cls, el: Json, list_type: ListType) -> "Entry":
        return cls(
            entry_type=list_type,
            id=int(el["id"]),
            title=el["title"],
            main_picture=el.get("main_picture", {}),
            alternative_titles=el["alternative_titles"],
            start_date=parse_date_safe(el.get("start_date")),
            end_date=parse_date_safe(el.get("end_date")),
            synopsis=el["synopsis"],
            mean=el.get("mean"),
            rank=el.get("rank"),
            popularity=int(el["popularity"]),
            num_list_users=int(el["num_list_users"]),
            num_scoring_users=int(el["num_scoring_users"]),
            nsfw=el["nsfw"],
            created_at=datetime.fromisoformat(el["created_at"]),
            updated_at=datetime.fromisoformat(el["updated_at"]),
            media_type=el["media_type"],
            status=el["status"],
            genres=IdInfo._parse_id_list(el, "genres"),
            list_status=el["my_list_status"],
            episode_count=el.get("num_episodes"),
            season=Season._parse(el.get("start_season")),
            source=el.get("source"),
            average_episode_duration=el.get("average_episode_duration"),
            rating=el.get("rating"),
            studios=IdInfo._parse_id_list(el, "studios"),
        )


def iter_api_list(json_file: PathIsh, list_type: ListType) -> Iterator[Entry]:
    data = json.loads(_expand_file(json_file).read_text())
    for el in data:
        yield Entry._parse(el, list_type)



from typing import NamedTuple, Optional, Union, Any, Dict, List
from datetime import date


import lxml.etree as ET  # type: ignore[import]

from ..paths import _expand_file, PathIsh
from ..list_type import ListType
from .common import parse_date_safe, strtobool

# hmm.. cant figure out the types for this
XMLElement = Any

Info = Dict[str, Union[int, str]]


# TODO: some of these are None if the text is empty, not sure how to mark those
class AnimeXML(NamedTuple):
    anime_id: int
    title: str
    media_type: str
    episodes: int
    my_id: int
    watched_episodes: int
    start_date: Optional[date]
    finish_date: Optional[date]
    rated: str
    score: int
    storage: str
    storage_value: float
    status: str
    comments: str
    times_watched: int
    rewatch_value: str
    priority: str
    tags: str
    rewatching: bool
    rewatching_ep: int
    discuss: bool
    sns: str
    update_on_import: bool

    @property
    def id(self) -> int:
        return self.anime_id

    @classmethod
    def _parse(cls, anime_el: XMLElement) -> "AnimeXML":
        return cls(
            anime_id=int(anime_el.find("series_animedb_id").text),
            title=anime_el.find("series_title").text,
            media_type=anime_el.find("series_type").text,
            episodes=int(anime_el.find("series_episodes").text),
            my_id=int(anime_el.find("my_id").text),
            watched_episodes=int(anime_el.find("my_watched_episodes").text),
            start_date=parse_date_safe(anime_el.find("my_start_date").text),
            finish_date=parse_date_safe(anime_el.find("my_finish_date").text),
            rated=anime_el.find("my_rated").text,
            score=int(anime_el.find("my_score").text),
            storage=anime_el.find("my_storage").text,
            storage_value=float(anime_el.find("my_storage_value").text),
            status=anime_el.find("my_status").text,
            comments=anime_el.find("my_comments").text,
            times_watched=int(anime_el.find("my_times_watched").text),
            rewatch_value=anime_el.find("my_rewatch_value").text,
            priority=anime_el.find("my_priority").text,
            tags=anime_el.find("my_tags").text,
            rewatching=strtobool(anime_el.find("my_rewatching").text.lower()),
            rewatching_ep=int(anime_el.find("my_rewatching_ep").text),
            discuss=strtobool(anime_el.find("my_discuss").text.lower()),
            sns=anime_el.find("my_sns").text,
            update_on_import=strtobool(anime_el.find("update_on_import").text),
        )


class MangaXML(NamedTuple):
    manga_id: int
    title: str
    volumes: int
    chapters: int
    my_id: int
    read_volumes: int
    read_chapters: int
    start_date: Optional[date]
    finish_date: Optional[date]
    scanlation_group: str
    score: int
    storage: str
    retail_volumes: int
    status: str
    comments: str
    times_read: int
    tags: str
    priority: str
    reread_value: str
    rereading: bool
    discuss: bool
    sns: str
    update_on_import: bool

    @property
    def id(self) -> int:
        return self.manga_id

    @classmethod
    def _parse(cls, manga_el: XMLElement) -> "MangaXML":
        return cls(
            manga_id=int(manga_el.find("manga_mangadb_id").text),
            title=manga_el.find("manga_title").text,
            volumes=int(manga_el.find("manga_volumes").text),
            chapters=int(manga_el.find("manga_chapters").text),
            my_id=int(manga_el.find("my_id").text),
            read_volumes=int(manga_el.find("my_read_volumes").text),
            read_chapters=int(manga_el.find("my_read_chapters").text),
            start_date=parse_date_safe(manga_el.find("my_start_date").text),
            finish_date=parse_date_safe(manga_el.find("my_finish_date").text),
            scanlation_group=manga_el.find("my_scanalation_group").text,
            score=int(manga_el.find("my_score").text),
            storage=manga_el.find("my_storage").text,
            retail_volumes=int(manga_el.find("my_retail_volumes").text),
            status=manga_el.find("my_status").text,
            comments=manga_el.find("my_comments").text,
            times_read=int(manga_el.find("my_times_read").text),
            tags=manga_el.find("my_tags").text,
            priority=manga_el.find("my_priority").text,
            reread_value=manga_el.find("my_reread_value").text,
            rereading=strtobool(manga_el.find("my_rereading").text.lower()),
            discuss=strtobool(manga_el.find("my_discuss").text.lower()),
            sns=manga_el.find("my_sns").text,
            update_on_import=strtobool(manga_el.find("update_on_import").text),
        )


Entry = Union[AnimeXML, MangaXML]


class XMLExport(NamedTuple):
    list_type: str
    info: Info
    entries: List[Entry]

    @staticmethod
    def _parse_info(info: XMLElement) -> Info:
        data: Info = {}
        for el in list(info):
            if str(el.text).isdigit():
                data[el.tag] = int(el.text)
            else:
                data[el.tag] = str(el.text)
        return data

    @classmethod
    def parse(cls, xml_file: str) -> "XMLExport":
        tree = ET.parse(xml_file)
        root = tree.getroot()
        info = cls._parse_info(root.find("myinfo"))
        export_type = int(info["user_export_type"])
        list_type: ListType
        entries: List[Entry] = []
        if export_type == 1:
            list_type = ListType.ANIME
            entries = [AnimeXML._parse(el) for el in root.findall("anime")]
        else:
            list_type = ListType.MANGA
            entries = [MangaXML._parse(el) for el in root.findall("manga")]

        return cls(info=info, entries=entries, list_type=list_type.value.lower())


def parse_xml(xml_file: PathIsh) -> XMLExport:
    return XMLExport.parse(str(_expand_file(xml_file)))



"""
Uses MALs API to download forum posts
"""

import json
from typing import Iterator

from .mal_session import MalSession
from ..paths import LocalDir, _expand_path
from ..common import Json, serialize


# one is created by, one is commented on, doesn't really matter which is which
FORUM_BASES = [
    "https://api.myanimelist.net/v2/forum/topics?user_name={mal_username}&limit=100",
    "https://api.myanimelist.net/v2/forum/topics?topic_user_name={mal_username}&limit=100",
]

FORUM_POST = "https://api.myanimelist.net/v2/forum/topic/{forum_id}?limit=100"


class ForumManager:
    """
    Download any forum posts which you've created/commented on
    """

    def __init__(self, localdir: LocalDir, mal_session: MalSession) -> None:
        self.localdir = localdir
        self.mal_session = mal_session
        self.mal_session.authenticate()
        self.forum_index_path = (
            _expand_path(self.localdir.data_dir / "forum") / "index.json"
        )

    def download_forum_index(self) -> Iterator[Json]:
        """
        Download a 'forum index' (i.e., the IDs/modification time for any
        post you've created/commented on) by paginating through the data
        """
        for forum_base in FORUM_BASES:
            url = forum_base.format(mal_username=self.localdir.username)
            for data_response in self.mal_session.paginate_all_data(url):
                yield from data_response

    def load_forum_index(self) -> Json:
        """
        Assuming the forum index exists, load the JSON file
        """
        assert self.forum_index_path.exists(), "Forum index doesnt exist!"
        return json.loads(self.forum_index_path.read_text())

    def update_forum_index(self) -> None:
        """
        Download and save the forum index. This doesn't attempt to
        do anything complicated, it just re-downloads the entire thing
        """
        data = list(self.download_forum_index())
        data_json = serialize(data)
        self.forum_index_path.write_text(data_json)

    def update_changed_forum_posts(self) -> None:
        """
        Load the forum index, and call .update_if_changed on each forum post
        """
        for forum_post in self.load_forum_index():
            forum = ForumPost(
                localdir=self.localdir,
                mal_session=self.mal_session,
                forum_id=int(forum_post["id"]),
                last_post_created_at=str(forum_post["last_post_created_at"]),
            )
            forum.update_if_changed()


class ForumPost:
    """
    Manages/Downloads data for one forum post
    """

    def __init__(
        self,
        localdir: LocalDir,
        mal_session: MalSession,
        forum_id: int,
        last_post_created_at: str,
    ) -> None:
        self.localdir = localdir
        self.mal_session = mal_session
        self.forum_id = forum_id
        self.last_post_created_at = last_post_created_at
        self.forum_path = (
            _expand_path(self.localdir.data_dir / "forum") / f"{forum_id}.json"
        )

    def forum_post_has_changed(self) -> bool:
        """
        Determine whether or not the forum post has new data, by looking at the last_post_created_at
        (downloaded in the forum index)
        """
        if self.forum_path.exists():
            data = json.loads(self.forum_path.read_text())
            if "last_post_created_at" in data:
                # if these aren't the same, the data has changed, should update
                return str(data["last_post_created_at"]) != self.last_post_created_at
            else:
                # hmm? didnt save last_post_created_at for some reason?
                return True
        else:
            # doesnt exist, need to download
            return True

    def download_forum_post(self) -> Json:
        """
        For a particular forum post, paginate through all the posts
        """
        responses = list(
            self.mal_session.paginate_all_data(
                FORUM_POST.format(forum_id=self.forum_id)
            )
        )
        assert len(responses) > 0, f"No data returned for {self.forum_id}!"
        # need to attach all 'posts' to the response
        data = responses[0]
        for resp in responses[1:]:
            data["posts"].extend(resp["posts"])
        # save when this was last updated from the forum index
        data["last_post_created_at"] = self.last_post_created_at
        return data

    def update_if_changed(self) -> None:
        """
        If the forum post has to be updated, update it
        """
        if not self.forum_post_has_changed():
            return
        data_json = json.dumps(self.download_forum_post())
        self.forum_path.write_text(data_json)



import os
import time
import warnings
import datetime
from typing import Any, Generator, Optional, Callable, Type, cast, Sequence
from urllib.parse import urlparse, parse_qs

import requests
import backoff  # type: ignore[import]
import simplejson

from .list_type import ListType

Json = Any

from malexport.log import logger

REQUEST_WAIT_TIME: int = int(os.environ.get("MALEXPORT_REQUEST_WAIT_TIME", 10))


def fibo_backoff() -> Generator[float, None, None]:
    """
    Fibonacci backoff, with the first 7 elements consumed.
    In other words, this starts at 13, 21, ....
    """
    fib = backoff.fibo()
    for _ in range(7):
        next(fib)
    for n in fib:
        yield float(n)


def backoff_hdlr(details: Any) -> None:
    warning_msg = "Backing off {wait:0.1f} seconds after {tries} tries with {args} {kwargs}".format(
        **details
    )
    warnings.warn(warning_msg)


@backoff.on_exception(
    fibo_backoff,
    cast(Sequence[Type[Exception]], (requests.RequestException,)),
    max_tries=3,
    on_backoff=backoff_hdlr,
)
def safe_request(
    url: str,
    *,
    method: str = "GET",
    session: Optional[requests.Session] = None,
    on_error: Optional[Callable[[requests.Response], Any]] = None,
    wait_time: int = REQUEST_WAIT_TIME,
    **kwargs: Any,
) -> requests.Response:
    """
    Sleep for a while, make a request, and retry 3 times if the request fails
    Can supply an on_error function to do some custom behaviour if theres an HTTP error
    """
    time.sleep(wait_time)
    sess: requests.Session
    if session is not None:
        sess = session
    else:
        sess = requests.Session()
    logger.info(f"Requesting {url}...")
    kwargs.setdefault("allow_redirects", True)
    r = sess.request(method, url, **kwargs)
    try:
        r.raise_for_status()
    except requests.RequestException as e:
        if on_error is not None:
            on_error(r)  # do something, e.g. refresh a expired bearer token
        raise e  # raise anyways, so this request retries
    return r


def safe_request_json(
    url: str, session: Optional[requests.Session] = None, **kwargs: Any
) -> Any:
    """
    Run a safe_request, then parse the response to JSON
    """
    return safe_request(url, session=session, **kwargs).json()


def default_encoder(o: Any) -> Any:
    if hasattr(o, "_asdict"):
        return o._asdict()
    if isinstance(o, ListType):
        return o.value
    if isinstance(o, datetime.datetime):
        return str(o)
    elif isinstance(o, datetime.date):
        return str(o)
    raise TypeError(f"{o} of type {type(o)} is not serializable")


def serialize(data: Any) -> str:
    try:
        import orjson  # type: ignore[import]

        bdata: bytes = orjson.dumps(data, default=default_encoder)
        return bdata.decode("utf-8")

    except ImportError:
        return simplejson.dumps(
            data,
            default=default_encoder,
            namedtuple_as_object=True,
        )


def extract_query_value(url: str, param: str) -> str:
    query_list = parse_qs(urlparse(url).query)[param]
    return query_list[0]



"""
Requests MAL Lists (animelist/mangalist) for a user, using MAL API
"""

from typing import List
from pathlib import Path

from ..list_type import ListType
from ..common import Json, serialize
from ..paths import LocalDir
from .mal_session import MalSession

BASE_URL = "https://api.myanimelist.net/v2/users/{username}/{list_type}list?limit=100&offset=0&nsfw=true&fields=id,title,main_picture,alternative_titles,start_date,end_date,synopsis,mean,rank,popularity,num_list_users,num_scoring_users,nsfw,created_at,updated_at,media_type,status,genres,my_list_status,num_episodes,start_season,broadcast,source,average_episode_duration,rating,pictures,background,related_anime,related_manga,recommendations,studios,statistics"


class APIList:
    """
    Requests/Updates the users data using the MAL API
    """

    def __init__(
        self, list_type: ListType, localdir: LocalDir, mal_session: MalSession
    ) -> None:
        self.localdir = localdir
        self.list_type = list_type
        self.mal_session = mal_session
        self.mal_session.authenticate()

    @property
    def list_path(self) -> Path:
        return self.localdir.data_dir / f"{self.list_type.value}list_api.json"

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(list_type={self.list_type}, localdir={self.localdir})"

    __str__ = __repr__

    def update_list(self) -> None:
        """
        Paginate through all the data from the MAL API
        """
        first_url = BASE_URL.format(
            list_type=self.list_type.value,
            username=self.localdir.username,
        )
        data: List[Json] = []
        for resp in self.mal_session.paginate_all_data(first_url):
            for entry in resp:
                data.append(entry["node"])
        encoded_data = serialize(data)
        self.list_path.write_text(encoded_data)



from typing import Optional

from ..paths import LocalDir
from ..list_type import ListType
from .mal_list import MalList
from .api_list import APIList
from .mal_session import MalSession
from .history import HistoryManager
from .forum import ForumManager
from .export_downloader import ExportDownloader
from .friends import FriendDownloader
from .messages import MessageDownloader


class Account:
    """
    This encapsulates everything with regards to exporting data from MAL
    """

    def __init__(self, localdir: LocalDir):
        self.localdir = localdir
        self.animelist = MalList(list_type=ListType.ANIME, localdir=self.localdir)
        self.mangalist = MalList(list_type=ListType.MANGA, localdir=self.localdir)
        self.animelist_api: Optional[APIList] = None
        self.mangalist_api: Optional[APIList] = None
        self.mal_session: Optional[MalSession] = None
        self.anime_episode_history: Optional[HistoryManager] = None
        self.manga_chapter_history: Optional[HistoryManager] = None
        self.forum_manager: Optional[ForumManager] = None
        self.export_downloader: Optional[ExportDownloader] = None
        self.friend_downloader: Optional[FriendDownloader] = None
        self.message_manager: Optional[MessageDownloader] = None

    def mal_api_authenticate(self) -> MalSession:
        """
        This authenticates the mal_session using the API
        If never done before, runs the OAuth flow. Else loads
        the access token from the config file
        """
        client_info = self.localdir.load_or_prompt_mal_client_info()
        self.mal_session = MalSession(
            client_id=client_info["client_id"], localdir=self.localdir
        )
        self.mal_session.authenticate()
        return self.mal_session

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(localdir={self.localdir})"

    __str__ = __repr__

    @staticmethod
    def from_username(username: str) -> "Account":
        """Alternate constructor to create an account from MAL username"""
        return Account(localdir=LocalDir.from_username(username))

    def update_lists(self, only: Optional[ListType] = None) -> None:
        """
        Uses the load.json endpoint to request anime/manga lists.
        Does not require any authentication
        """
        if only == ListType.ANIME or only is None:
            self.animelist.update_list()
        if only == ListType.MANGA or only is None:
            self.mangalist.update_list()

    def update_api_lists(self, only: Optional[ListType] = None) -> None:
        """
        Uses MALs API to request anime/manga lists
        Requires authentication, but includes more data than load.json
        """
        self.mal_api_authenticate()
        assert self.mal_session is not None
        if self.animelist_api is None:
            self.animelist_api = APIList(
                list_type=ListType.ANIME,
                localdir=self.localdir,
                mal_session=self.mal_session,
            )
        if self.mangalist_api is None:
            self.mangalist_api = APIList(
                list_type=ListType.MANGA,
                localdir=self.localdir,
                mal_session=self.mal_session,
            )
        if only == ListType.ANIME or only is None:
            self.animelist_api.update_list()
        if only == ListType.MANGA or only is None:
            self.mangalist_api.update_list()

    def update_exports(self) -> None:
        """
        Uses selenium to export/extract the XML files from MAL.
        Requires authentication (MAL Username/Password)
        """
        self.export_downloader = ExportDownloader(self.localdir)
        self.export_downloader.export_lists()

    def update_history(
        self,
        only: Optional[ListType] = None,
        count: Optional[int] = None,
        driver_type: str = "chrome",
        use_merged_file: bool = False,
    ) -> None:
        """
        Uses selenium to download episode/chapter history one entry at a time.

        This takes quite a while, and requires authentication (MAL Username/Password)
        If count is specified, only requests the first 'count' IDs found in your history
        """
        self.anime_episode_history = HistoryManager(
            list_type=ListType.ANIME,
            localdir=self.localdir,
            driver_type=driver_type,
            use_merged_file=use_merged_file,
        )
        self.manga_chapter_history = HistoryManager(
            list_type=ListType.MANGA,
            localdir=self.localdir,
            driver_type=driver_type,
            use_merged_file=use_merged_file,
        )
        if only == ListType.ANIME or only is None:
            self.anime_episode_history.update_history(count=count)
        if only == ListType.MANGA or only is None:
            self.manga_chapter_history.update_history(count=count)

    def update_messages(
        self, start_page: int = 1, thread_count: Optional[int] = None
    ) -> None:
        """
        Download/Update DMs for your account
        """
        self.message_manager = MessageDownloader(
            self.localdir,
            till_same_limit=thread_count,
        )
        self.message_manager.update_messages(start_page=start_page)

    def update_forum_posts(self) -> None:
        """
        Uses the MAL API to download any forum posts which you've created/commented on

        Requires you to go to https://myanimelist.net/apiconfig and create a Client. You can
        use any App Type other than Web, this doesn't use a Client Secret

        You can use http://localhost as the App Redirect URL
        """
        self.mal_api_authenticate()
        assert self.mal_session is not None
        self.forum_manager = ForumManager(
            localdir=self.localdir, mal_session=self.mal_session
        )
        self.forum_manager.update_forum_index()
        self.forum_manager.update_changed_forum_posts()

    def update_friends(self) -> None:
        """
        Uses Jikan to download your friends from MAL
        """
        self.friend_downloader = FriendDownloader(localdir=self.localdir)
        self.friend_downloader.update_friend_index()



import os
import sys
from pathlib import Path
from typing import Callable, Optional, Any

import click

from .list_type import ListType
from .paths import default_zip_base


@click.group()
def main() -> None:
    """
    exports your data from MyAnimeList
    """


# shared click options/args
USERNAME = click.option(
    "-u",
    "--username",
    "username",
    required=True,
    help="Username to use",
    envvar="MAL_USERNAME",
)

ONLY = click.option(
    "-o",
    "--only",
    type=click.Choice(["anime", "manga"], case_sensitive=False),
    required=False,
    help="Only update anime or manga history specifically",
)

STREAM = click.option(
    "-s",
    "--stream",
    is_flag=True,
    default=False,
    help="Stream JSON objects instead of printing a list",
)


def apply_shared(*chosen: Any) -> Any:
    """
    Decorator to apply the username argument
    """

    def _add_options(func: Callable[..., None]) -> Callable[..., None]:
        for dec in chosen:
            func = dec(func)
        return func

    return _add_options


@main.group()
def update() -> None:
    """
    update data for an account
    """


@update.command(name="all", short_help="update all data")
@apply_shared(USERNAME)
def _all(username: str) -> None:
    """
    update all data for the account
    """
    from .exporter import Account

    acc = Account.from_username(username)
    acc.update_lists()
    acc.update_api_lists()
    acc.update_forum_posts()
    acc.update_history()
    acc.update_friends()
    acc.update_exports()
    acc.update_messages()


@update.command(name="lists", short_help="update animelist and mangalists")
@apply_shared(USERNAME, ONLY)
def _lists_update(only: str, username: str) -> None:
    from .exporter import Account

    acc = Account.from_username(username)
    only_update: Optional[ListType] = None
    if only is not None:
        only_update = ListType.__members__[only.upper()]
    acc.update_lists(only=only_update)


@update.command(name="messages", short_help="update messages (DMs)")
@apply_shared(USERNAME)
@click.option(
    "--thread-count",
    type=int,
    default=None,
    help="how many new threads to update before giving up",
)
@click.option(
    "--start-page", type=int, default=1, help="which page to start requesting from"
)
def _messages_update(
    username: str, start_page: int, thread_count: Optional[int] = None
) -> None:
    from .exporter import Account

    acc = Account.from_username(username)
    acc.update_messages(start_page=start_page, thread_count=thread_count)


@update.command(
    name="api-lists", short_help="update animelist and mangalists using the API"
)
@apply_shared(USERNAME, ONLY)
def _api_lists(only: str, username: str) -> None:
    from .exporter import Account

    acc = Account.from_username(username)
    only_update: Optional[ListType] = None
    if only is not None:
        only_update = ListType.__members__[only.upper()]
    acc.update_api_lists(only=only_update)


@update.command(name="export", short_help="export xml lists")
@apply_shared(USERNAME)
def _export(username: str) -> None:
    from .exporter import Account

    acc = Account.from_username(username)
    acc.update_exports()


@update.command(name="history", short_help="update episode history")
@apply_shared(USERNAME, ONLY)
@click.option(
    "-c",
    "--count",
    "count",
    default=None,
    type=int,
    help="Only request the first 'count' entries in the users episode history",
)
@click.option(
    "--driver-type",
    default="chrome",
    type=click.Choice(["chrome", "firefox"]),
    help="whether to use chromedriver/geckodriver as the selenium browser",
)
@click.option(
    "--use-merged-file",
    default=False,
    is_flag=True,
    envvar="MALEXPORT_USE_MERGED_FILE",
    help="use a single merged JSON file instead of storing history data in individual files",
)
def _history(
    username: str,
    only: Optional[str],
    driver_type: str,
    count: Optional[int],
    use_merged_file: bool,
) -> None:
    from .exporter import Account

    acc = Account.from_username(username)
    only_update: Optional[ListType] = None
    if only is not None:
        only_update = ListType.__members__[only.upper()]
    acc.update_history(
        only=only_update,
        count=count,
        driver_type=driver_type,
        use_merged_file=use_merged_file,
    )


@update.command(name="forum", short_help="update forum posts")
@apply_shared(USERNAME)
def _forum(username: str) -> None:
    from .exporter import Account

    acc = Account.from_username(username)
    acc.update_forum_posts()


@update.command(name="friends", short_help="update friends")
@apply_shared(USERNAME)
def _friends(username: str) -> None:
    from .exporter import Account

    acc = Account.from_username(username)
    acc.update_friends()


@main.group(name="parse")
def parse() -> None:
    """parse the resulting exported files"""


@parse.command(name="xml", short_help="parse the XML export files")
@click.argument("XML_FILE")
def _xml(xml_file: str) -> None:
    from .parse import parse_xml
    from .common import serialize

    xml_data = parse_xml(xml_file)
    click.echo(serialize(xml_data))


@parse.command(name="list", short_help="parse the list file")
@click.option(
    "--type",
    "_type",
    type=click.Choice(["anime", "manga"], case_sensitive=False),
    required=False,
    help="Specify type of list. If not supplied, this tries to guess based on the filename",
)
@apply_shared(STREAM)
@click.argument("LIST_FILE")
def _list_parse(_type: Optional[str], list_file: str, stream: bool) -> None:
    from .parse import parse_list
    from .common import serialize

    chosen_type: ListType
    if _type is not None:
        chosen_type = ListType.__members__[_type.upper()]
    else:
        # infer type
        chosen_type = (
            ListType.ANIME if "anime" in os.path.basename(list_file) else ListType.MANGA
        )
    idata = parse_list(list_file, list_type=chosen_type)
    if stream:
        for i in idata:
            sys.stdout.write(serialize(i))
        sys.stdout.flush()
    else:
        click.echo(serialize(list(idata)))


@parse.command(name="api-list", short_help="parse the API list file")
@click.option(
    "--type",
    "_type",
    type=click.Choice(["anime", "manga"], case_sensitive=False),
    required=False,
    help="Specify type of list. If not supplied, this tries to guess based on the filename",
)
@apply_shared(STREAM)
@click.argument("API_LIST_FILE")
def _api_list_file(_type: Optional[str], api_list_file: str, stream: bool) -> None:
    from .parse import iter_api_list
    from .common import serialize

    chosen_type: ListType
    if _type is not None:
        chosen_type = ListType.__members__[_type.upper()]
    else:
        # infer type
        chosen_type = (
            ListType.ANIME
            if "anime" in os.path.basename(api_list_file)
            else ListType.MANGA
        )
    idata = iter_api_list(api_list_file, list_type=chosen_type)
    if stream:
        for i in idata:
            sys.stdout.write(serialize(i))
        sys.stdout.flush()
    else:
        click.echo(serialize(list(idata)))


@parse.command(name="forum", short_help="extract forum posts by your user")
@apply_shared(USERNAME)
def _forum_parse(username: str) -> None:
    from .parse import iter_forum_posts
    from .common import serialize

    click.echo(serialize(list(iter_forum_posts(username))))


@parse.command(
    name="combine", short_help="combines lists, api-lists, xml and history data"
)
@apply_shared(USERNAME, ONLY)
def _combine_parse(only: Optional[str], username: str) -> None:
    """
    This combines relevant info from the lists, xml and history files
    It removes some of the commonly unused fields, and uses the xml for rewatch info/better dates

    It doesn't require you have a list export
    """
    from .parse.combine import combine
    from .common import serialize

    anime, manga = combine(username)
    if only == "anime":
        click.echo(serialize(anime))
    elif only == "manga":
        click.echo(serialize(manga))
    else:
        click.echo(serialize({"anime": anime, "manga": manga}))


@parse.command(name="history", short_help="parse downloaded user history")
@apply_shared(USERNAME)
def _history_parse(username: str) -> None:
    from .parse import iter_user_history
    from .common import serialize

    click.echo(serialize(list(iter_user_history(username))))


@parse.command(name="messages", short_help="parse downloaded message history")
@apply_shared(USERNAME)
def _messages_parse(username: str) -> None:
    from .parse import iter_user_threads
    from .common import serialize

    click.echo(serialize(list(iter_user_threads(username))))


@parse.command(name="friends", short_help="parse user friends")
@apply_shared(USERNAME)
def _friends_parse(username: str) -> None:
    from .common import serialize
    from .parse import iter_friends

    click.echo(serialize(list(iter_friends(username))))


@main.group(short_help="recover data for deleted MAL entries")
def recover_deleted() -> None:
    """
    lets you recover data from zip backups of your malexport dir
    """


_ISO_FORMAT = "%Y-%m-%dT%H-%M-%SZ"


def utcnow() -> str:
    from datetime import datetime, timezone

    return datetime.now(timezone.utc).strftime(_ISO_FORMAT)


@recover_deleted.command(short_help="zips your current malexport dir")
@apply_shared(USERNAME)
def backup(username: str) -> None:
    import shutil

    from .paths import default_zip_base
    from .exporter import Account

    backup_to_dir = default_zip_base / username
    backup_to_dir.mkdir(parents=True, exist_ok=True)
    from_dir = Account.from_username(username).localdir.data_dir

    backup_zip_base = str(backup_to_dir / f"{utcnow()}")
    backup_zip_full = f"{backup_zip_base}.zip"

    # shutil doesnt want the '.zip' at the end of the file
    # automatically compresses if possible
    shutil.make_archive(backup_zip_base, "zip", root_dir=from_dir, base_dir=".")

    click.echo(f"Backed up {from_dir} to {backup_zip_full}", err=True)
    click.echo(
        "Backup Size: {:.2f} MB".format(
            Path(backup_zip_full).stat().st_size / 1024**2
        ),
        err=True,
    )


@recover_deleted.command(short_help="stats about mal-id-cache git dir")
def approved_ids_stats() -> None:
    from .parse.recover_deleted_entries import Approved

    Approved.git_pull()
    apr = Approved.parse_from_git_dir()

    click.echo(f"Approved Anime: {len(apr.anime)}")
    click.echo(f"Approved Manga: {len(apr.manga)}")


@recover_deleted.command(short_help="update approved ids")
def approved_update() -> None:
    from .parse.recover_deleted_entries import Approved

    commit_id = Approved.git_pull()

    click.echo(f"Updated mal-id-cache to commit {commit_id}")


@recover_deleted.command(short_help="recover deleted MAL entries from your zip backups")
@click.option(
    "--data-dir",
    type=click.Path(exists=True, file_okay=False, path_type=Path),
    default=str(default_zip_base),
)
@click.option(
    "-F",
    "--filter-with-activity",
    is_flag=True,
    help="only return items which have some activity (read/watched)",
)
@apply_shared(ONLY)
def recover(data_dir: Path, filter_with_activity: bool, only: str) -> None:
    from .parse.recover_deleted_entries import recover_deleted as rec_del, Approved
    from .common import serialize

    full_resp = {}

    for acc in data_dir.iterdir():
        username = acc.name
        zips = sorted(acc.glob("*.zip"), key=lambda x: x.name)

        rec_anime, rec_manga = rec_del(
            approved=Approved.parse_from_git_dir(),
            username=username,
            backups=zips,
            filter_with_activity=filter_with_activity,
        )

        resp: Any = {
            "anime": rec_anime,
            "manga": rec_manga,
        }

        # if user specified only, only return that type
        if only:
            assert only in resp
            resp = resp[only]
            assert isinstance(resp, list)

        full_resp[username] = resp
    click.echo(serialize(full_resp))


if __name__ == "__main__":
    main(prog_name="malexport")



"""
Exports your list as XML from MAL using Selenium
"""

import os
import time
import shutil
import gzip
from typing import List, Optional

from selenium.webdriver.support.ui import WebDriverWait  # type: ignore[import]
from selenium.webdriver.common.by import By  # type: ignore[import]
from selenium.webdriver.support import expected_conditions as EC  # type: ignore[import]
from selenium.common.exceptions import TimeoutException, WebDriverException  # type: ignore[import]

from .driver import webdriver, driver_login, wait, TEMP_DOWNLOAD_DIR, Browser
from ..list_type import ListType
from ..paths import LocalDir
from ..log import logger

TRY_EXPORT_TIMES = int(os.environ.get("MALEXPORT_EXPORT_TRIES", 3))

EXPORT_PAGE = "https://myanimelist.net/panel.php?go=export"
EXPORT_BUTTON_CSS = "input[value='Export My List']"
DOWNLOAD_BUTTON = ".goodresult>a"


class ExportDownloader:
    """
    Downloads the XML exports for your account
    """

    def __init__(self, localdir: LocalDir) -> None:
        self.localdir = localdir
        self.animelist_path = self.localdir.data_dir / "animelist.xml"
        self.mangalist_path = self.localdir.data_dir / "mangalist.xml"
        self._driver: Optional[Browser] = None

    @property
    def driver(self) -> Browser:
        if self._driver is None:
            self._driver = webdriver(browser_type="chrome")
        return self._driver

    def authenticate(self) -> None:
        """Logs in to MAL using your MAL username/password"""
        # requires chrome because uses experimental flag to save downloads to custom dir
        driver_login(self.driver, self.localdir)

    def export_lists(self) -> None:
        """Exports the anime/manga lists, then extracts the gz files into the data dir"""
        self.authenticate()
        self.export_with_retry(ListType.ANIME)
        self.export_with_retry(ListType.MANGA)
        self.extract_gz_files()

    def export_with_retry(self, list_type: ListType, *, times: int = 0) -> None:
        """
        If exporting a list fails, resets the browser and tries again
        """
        try:
            self.export_list(list_type)
        except WebDriverException as e:
            times += 1
            logger.exception(
                f"Failed to export {list_type.value}, retrying ({times} of {TRY_EXPORT_TIMES})",
                exc_info=e,
            )
            if times > TRY_EXPORT_TIMES:
                return
            self.authenticate()
            # if user manually accepted/a file already present, skip retry
            if len(self._list_files(list_type=list_type)) > 0:
                logger.info("Found downloaded file, skipping retry...")
                time.sleep(1)
                return
            self.export_with_retry(list_type, times=times)  # recursive call

    def export_list(self, list_type: ListType) -> None:
        """
        Exports a particular list types' XML file, waits a while so that it can finish downloading
        The only difference between anime and manga is what is selected in the dialog
        """
        time.sleep(1)
        logger.info(f"Downloading {list_type.value} export")
        if self.driver.current_url != EXPORT_PAGE:
            self.driver.get(EXPORT_PAGE)
        export_button_selector = tuple([By.CSS_SELECTOR, EXPORT_BUTTON_CSS])
        WebDriverWait(self.driver, 15).until(  # type: ignore[no-untyped-call]
            EC.visibility_of_element_located(export_button_selector)  # type: ignore[no-untyped-call]
        )
        if list_type == ListType.MANGA:
            self.driver.execute_script("""$("#dialog select.inputtext").val(2)""")  # type: ignore[no-untyped-call]
        self.driver.find_element(By.CSS_SELECTOR, EXPORT_BUTTON_CSS).click()  # type: ignore[no-untyped-call]
        time.sleep(0.25)
        try:
            WebDriverWait(self.driver, 5).until(EC.alert_is_present())  # type: ignore[no-untyped-call]
        except TimeoutException:
            pass
        alert = self.driver.switch_to.alert
        time.sleep(0.25)
        alert.accept()  # type: ignore[no-untyped-call]
        time.sleep(0.25)
        download_button_selector = tuple([By.CSS_SELECTOR, DOWNLOAD_BUTTON])
        try:
            # hmm -- this page seems to be there sometimes, but not others?
            WebDriverWait(self.driver, 5).until(  # type: ignore[no-untyped-call]
                EC.element_to_be_clickable(download_button_selector)  # type: ignore[no-untyped-call]
            )
            self.driver.find_element(By.CSS_SELECTOR, DOWNLOAD_BUTTON).click()
        except TimeoutException:
            pass
        logger.debug("Waiting for download...")
        wait()

    def _list_files(
        self, path: str = TEMP_DOWNLOAD_DIR, list_type: Optional[ListType] = None
    ) -> List[str]:
        """
        List files in the temporary download directory, warn if there
        are multiple files/partially downloaded files
        """
        files = os.listdir(path)
        animelist_gzs = [
            f for f in files if f.startswith("animelist_") and f.endswith(".gz")
        ]
        mangalist_gzs = [
            f for f in files if f.startswith("mangalist_") and f.endswith(".gz")
        ]
        archive_files = animelist_gzs + mangalist_gzs
        if list_type == ListType.ANIME:
            archive_files = animelist_gzs
        elif list_type == ListType.MANGA:
            archive_files = mangalist_gzs
        logger.debug(archive_files)
        return archive_files

    def extract_gz_files(self) -> None:
        """
        Wait till two files (the anime/manga gz files) exist in the temporary download
        directory, then extract them to the data directory
        """
        while (
            len(self._list_files(list_type=ListType.ANIME)) < 1
            and len(self._list_files(list_type=ListType.MANGA)) < 1
        ):
            logger.info(
                f"Waiting till anime/manga list export files exist, currently {os.listdir(TEMP_DOWNLOAD_DIR)}"
            )
            time.sleep(0.5)

        for (archive_name, target) in zip(
            [
                self._list_files(list_type=ListType.ANIME)[0],
                self._list_files(list_type=ListType.MANGA)[0],
            ],
            [self.animelist_path, self.mangalist_path],
        ):
            archive_path = os.path.join(TEMP_DOWNLOAD_DIR, archive_name)
            logger.info(f"Extracting {archive_path} to {target}")
            with gzip.open(archive_path, "rb") as gz_in:
                with target.open("wb") as f:
                    shutil.copyfileobj(gz_in, f)



import json
from typing import NamedTuple, Union, List, Optional, TypeVar, Iterator
from datetime import date

from .common import strtobool, parse_short_date
from ..list_type import ListType
from ..common import Json
from ..paths import PathIsh, _expand_file

T = TypeVar("T")


def filter_none(lst: List[Optional[T]]) -> List[T]:
    lst_new: List[T] = []
    for x in lst:
        if x is not None:
            lst_new.append(x)
    return lst_new


class IdInfo(NamedTuple):
    id: int
    name: str

    @staticmethod
    def _parse(data: Optional[Json]) -> Optional["IdInfo"]:
        if isinstance(data, dict) and "id" in data and "name" in data:
            return IdInfo(id=data["id"], name=data["name"])
        else:
            return None

    @classmethod
    def _parse_id_list(cls, el: Json, key: str) -> List["IdInfo"]:
        return filter_none([cls._parse(e) for e in list(el.get(key) or [])])


class Season(NamedTuple):
    year: int
    season: str

    @staticmethod
    def _parse(season_data: Optional[Json]) -> Optional["Season"]:
        if (
            isinstance(season_data, dict)
            and "year" in season_data
            and "season" in season_data
        ):
            return Season(year=season_data["year"], season=season_data["season"])
        else:
            return None


ANIME_STATUS_MAP = {
    1: "Currently Watching",
    2: "Completed",
    3: "On Hold",
    4: "Dropped",
    6: "Plan to Watch",
}

MANGA_STATUS_MAP = {
    1: "Currently Reading",
    2: "Completed",
    3: "On Hold",
    4: "Dropped",
    6: "Plan to Read",
}

ANIME_AIRING_STATUS_MAP = {
    1: "Currently Airing",
    2: "Finished Airing",
    3: "Not Yet Aired",
}

MANGA_PUBLISHING_STATUS_MAP = {
    1: "Currently Publishing",
    2: "Finished Publishing",
    3: "Not Yet Published",
    4: "On Hiatus",
    5: "Discontinued",
}


class AnimeEntry(NamedTuple):
    status: str
    score: int
    tags: str
    rewatching: bool
    watched_episodes: int
    title: str
    episodes: int
    airing_status: str
    id: int
    studios: List[IdInfo]
    licensors: List[IdInfo]
    genres: List[IdInfo]
    demographics: List[IdInfo]
    season: Optional[Season]
    has_episode_video: bool
    has_promotion_video: bool
    has_video: bool
    video_url: str
    url: str
    image_path: str
    is_added_to_list: bool
    media_type: str
    rating: str
    start_date: Optional[date]
    finish_date: Optional[date]
    air_start_date: Optional[date]
    air_end_date: Optional[date]
    days: Optional[int]
    storage: str
    priority: str

    @classmethod
    def _parse(cls, el: Json) -> "AnimeEntry":
        return cls(
            status=ANIME_STATUS_MAP[el["status"]],
            score=el["score"],
            tags=el["tags"],
            rewatching=strtobool(el["is_rewatching"]),
            watched_episodes=el["num_watched_episodes"],
            title=el["anime_title"],
            episodes=el["anime_num_episodes"],
            airing_status=ANIME_AIRING_STATUS_MAP[el["anime_airing_status"]],
            id=el["anime_id"],
            studios=IdInfo._parse_id_list(el, "anime_studios"),
            licensors=IdInfo._parse_id_list(el, "anime_licensors"),
            genres=IdInfo._parse_id_list(el, "genres"),
            demographics=IdInfo._parse_id_list(el, "demographics"),
            season=Season._parse(el["anime_season"]),
            has_video=el["has_video"],
            video_url=el["video_url"],
            has_episode_video=el["has_episode_video"],
            has_promotion_video=el["has_promotion_video"],
            url=el["anime_url"],
            image_path=el["anime_image_path"],
            is_added_to_list=el["is_added_to_list"],
            media_type=el["anime_media_type_string"],
            rating=el["anime_mpaa_rating_string"],
            start_date=parse_short_date(el["start_date_string"] or ""),
            finish_date=parse_short_date(el["finish_date_string"] or ""),
            air_start_date=parse_short_date(el["anime_start_date_string"]),
            air_end_date=parse_short_date(el["anime_end_date_string"]),
            days=el["days_string"],
            storage=el["storage_string"],
            priority=el["priority_string"],
        )


class MangaEntry(NamedTuple):
    status: str
    score: int
    tags: str
    rereading: bool
    read_chapters: int
    read_volumes: int
    title: str
    chapters: int
    volumes: int
    publishing_status: str
    id: int
    genres: List[IdInfo]
    demographics: List[IdInfo]
    manga_magazines: List[IdInfo]
    url: str
    image_path: str
    is_added_to_list: bool
    media_type: str
    start_date: Optional[date]
    finish_date: Optional[date]
    publish_start_date: Optional[date]
    publish_end_date: Optional[date]
    days: Optional[int]
    retail: str
    priority: str

    @classmethod
    def _parse(cls, el: Json) -> "MangaEntry":
        return cls(
            status=MANGA_STATUS_MAP[el["status"]],
            score=el["score"],
            tags=el["tags"],
            rereading=strtobool(el["is_rereading"]),
            read_chapters=el["num_read_chapters"],
            read_volumes=el["num_read_volumes"],
            title=el["manga_title"],
            chapters=el["manga_num_chapters"],
            volumes=el["manga_num_volumes"],
            publishing_status=MANGA_PUBLISHING_STATUS_MAP[
                el["manga_publishing_status"]
            ],
            id=el["manga_id"],
            manga_magazines=IdInfo._parse_id_list(el, "manga_magazines"),
            genres=IdInfo._parse_id_list(el, "genres"),
            demographics=IdInfo._parse_id_list(el, "demographics"),
            url=el["manga_url"],
            image_path=el["manga_image_path"],
            is_added_to_list=el["is_added_to_list"],
            media_type=el["manga_media_type_string"],
            start_date=parse_short_date(el["start_date_string"] or ""),
            finish_date=parse_short_date(el["finish_date_string"] or ""),
            publish_start_date=parse_short_date(el["manga_start_date_string"]),
            publish_end_date=parse_short_date(el["manga_end_date_string"]),
            days=el["days_string"],
            retail=el["retail_string"],
            priority=el["priority_string"],
        )


Entry = Union[AnimeEntry, MangaEntry]


def iter_user_list(json_file: str, list_type: ListType) -> Iterator[Entry]:
    with open(json_file) as f:
        data = json.load(f)
    if list_type == ListType.ANIME:
        for el in data:
            yield AnimeEntry._parse(el)
    else:
        for el in data:
            yield MangaEntry._parse(el)


def parse_file(json_file: PathIsh, list_type: ListType) -> Iterator[Entry]:
    yield from iter_user_list(str(_expand_file(json_file)), list_type)



"""
Requests MAL Lists (animelist/mangalist) for a user,
doesn't require authentication

This requests using the load.json endpoint, the same
mechanism that works on modern MAL lists
"""

import os
import json
from typing import List
from pathlib import Path

import requests

from ..list_type import ListType
from ..common import Json, safe_request_json, logger, serialize
from ..paths import LocalDir

# this is order=5, which requests items that were edited by you recently
BASE_URL = "https://myanimelist.net/{list_type}list/{username}/load.json?status=7&order=5&offset={offset}"

LIST_USER_AGENT = os.environ.get(
    "MALEXPORT_LIST_USER_AGENT",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:88.0) Gecko/20100101 Firefox/88.0",
)


OFFSET_CHUNK = 300


def handle_unauthorized(r: requests.Response) -> None:
    if r.status_code in [400, 403]:
        if (
            "Content-Type" in r.headers
            and "application/json" in r.headers["Content-Type"]
        ):
            logger.warning(r.json())
        else:
            logger.warning(r.text)
        raise RuntimeError(
            f"Auth/Permission error retrieving {r.url}, probably a permission error; the user has restricted access to their list"
        )


class MalList:
    """
    Requests/Updates the load.json endpoint for a particular user and list type
    """

    def __init__(self, list_type: ListType, localdir: LocalDir):
        self.list_type = list_type
        self.localdir = localdir

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(list_type={self.list_type}, localdir={self.localdir})"

    __str__ = __repr__

    def offset_url(self, offset: int) -> str:
        """
        Creates the offset URL (ordered by most recently edited) for a particular type/username/offset
        This is the same request that runs when you scroll down on a modern list and the next 300 chunks
        of entries load in
        """
        return BASE_URL.format(
            list_type=self.list_type.value,
            username=self.localdir.username,
            offset=offset,
        )

    @property
    def list_path(self) -> Path:
        return self.localdir.data_dir / f"{self.list_type.value}list.json"

    def load_list(self) -> List[Json]:
        """
        Load the list from the JSON file
        """
        if self.list_path.exists():
            try:
                return list(json.loads(self.list_path.read_text()))
            except json.JSONDecodeError:
                pass
        raise FileNotFoundError(f"No file found at {self.list_type.value}")

    def update_list(self) -> None:
        """
        Paginate through all the data till you hit a chunk of data which has
        less than OFFSET_CHUNK (300) items
        """
        list_data: List[Json] = []
        # overwrite the list with new data
        offset = 0
        session = requests.Session()
        session.headers.update({"User-Agent": LIST_USER_AGENT})
        while True:
            url = self.offset_url(offset)
            new_data = safe_request_json(
                url, session=session, on_error=handle_unauthorized
            )
            list_data.extend(new_data)
            if len(new_data) < OFFSET_CHUNK:
                logger.info(
                    f"After {offset // OFFSET_CHUNK} pages, only received {len(new_data)} (typical pages have {OFFSET_CHUNK}), stopping..."
                )
                break
            offset += OFFSET_CHUNK
        encoded_data = serialize(list_data)
        self.list_path.write_text(encoded_data)



import os
import re
from typing import Optional, Union, List
from datetime import date

from distutils.util import strtobool as strtoint


def split_tags(tags: str) -> List[str]:
    return list(re.split(r"\s*,\s*", tags.strip()))


def parse_date_safe(d: Optional[str]) -> Optional[date]:
    if d is None:
        return None
    if isinstance(d, str) and len(d.strip()) == 4 and d.isdigit():
        return date(year=int(d), month=1, day=1)
    try:
        return date.fromisoformat(d)
    except ValueError:  # no date supplied, uses '0000-00-00'
        return None


# this maintains a buffer of 5 years into the future to account for
# entries which haven't yet been released
# i.e., if CUTOFF_DATE is 1930
# '29' parses to 2029
# '30' parses to 1930
# '31' parses to 1931
CUTOFF_DATE = int(os.environ.get("MALEXPORT_CUTOFF_DATE", date.today().year + 5))

DATE_REGEX = re.compile(r"(\d+)-(\d+)-(\d+)")


def parse_short_date(d: Optional[str]) -> Optional[date]:
    """
    Parses dates that look like '30-06-73' or '04-09-20'
    Is not always 100% accurate because '04-09-20'
    could mean 1920 or 2020
    """
    if d is None:
        return None
    short_cutoff = int(str(CUTOFF_DATE)[-2:])
    m = DATE_REGEX.match(d)
    if m:
        day, month, year_short = [int(k) for k in m.groups()]
        if year_short < short_cutoff:
            year = 2000 + year_short
        else:
            year = 1900 + year_short
        # set defaults for items which don't have months/days
        if day == 0:
            day = 1
        if month == 0:
            month = 1
        try:
            return date(year=year, month=month, day=day)
        except ValueError:
            return None
    else:
        return None


def strtobool(val: Union[str, int, bool]) -> bool:
    if isinstance(val, bool):
        return val
    if isinstance(val, int):
        val = str(val)
    return bool(strtoint(val.lower()))



"""
A helper module to combine data from the list,
export and history modules into data I find useful
"""

import os
from typing import Dict, List, NamedTuple, Optional, TypeVar, Tuple, Union, Set
from pathlib import Path

from ..list_type import ListType
from ..log import logger
from ..paths import LocalDir
from .common import split_tags
from .history import History, HistoryEntry, iter_history_from_dir
from .mal_list import (
    AnimeEntry,
    MangaEntry,
    parse_file as parse_user_history,
)
from .api_list import iter_api_list, Entry
from .xml import AnimeXML, MangaXML, parse_xml


T = TypeVar("T")

FILTER_TAGS = "MALEXPORT_COMBINE_FILTER_TAGS"


class AnimeData(NamedTuple):
    XMLData: AnimeXML
    history: List[HistoryEntry]
    JSONList: Optional[AnimeEntry]
    APIList: Optional[Entry]
    username: str

    @property
    def id(self) -> int:
        return self.XMLData.id

    @property
    def tags_list(self) -> List[str]:
        if self.JSONList:
            return split_tags(self.JSONList.tags)
        return []


class MangaData(NamedTuple):
    XMLData: MangaXML
    history: List[HistoryEntry]
    JSONList: Optional[MangaEntry]
    APIList: Optional[Entry]
    username: str

    @property
    def id(self) -> int:
        return self.XMLData.id

    @property
    def tags_list(self) -> List[str]:
        if self.JSONList:
            return split_tags(self.JSONList.tags)
        return []


CombineResults = Tuple[List[AnimeData], List[MangaData]]


def combine(username: str, data_dir: Optional[Path] = None) -> CombineResults:
    if data_dir is None:
        acc = LocalDir.from_username(username)
        data_dir = acc.data_dir

    assert data_dir is not None

    # read history
    history = list(iter_history_from_dir(data_dir))
    anime_history: Dict[int, History] = {}
    manga_history: Dict[int, History] = {}
    for h in history:
        if h.list_type == "anime":
            # should never overwrite stuff by mistake
            assert h.mal_id not in anime_history
            anime_history[h.mal_id] = h
        else:
            assert h.mal_id not in manga_history
            manga_history[h.mal_id] = h

    # theres a possibility that the JSON exports don't
    # exist, because of private lists
    animelist_json_data: Dict[int, AnimeEntry] = {}
    mangalist_json_data: Dict[int, MangaEntry] = {}
    if (data_dir / "animelist.json").exists():
        animelist_json_data = {
            el.id: el  # type: ignore[union-attr,misc]
            for el in parse_user_history(
                str(data_dir / "animelist.json"), list_type=ListType.ANIME
            )
        }
    if (data_dir / "mangalist.json").exists():
        mangalist_json_data = {
            el.id: el  # type: ignore[union-attr,misc]
            for el in parse_user_history(
                str(data_dir / "mangalist.json"), list_type=ListType.MANGA
            )
        }

    # xml exports should always exist
    animelist_xml_data: Dict[int, AnimeXML] = {
        el.id: el  # type: ignore[union-attr,misc]
        for el in parse_xml(str(data_dir / "animelist.xml")).entries
    }
    mangalist_xml_data: Dict[int, MangaXML] = {
        el.id: el  # type: ignore[union-attr,misc]
        for el in parse_xml(str(data_dir / "mangalist.xml")).entries
    }

    # list using the API
    animelist_api_json_data: Dict[int, Entry] = {}
    mangalist_api_json_data: Dict[int, Entry] = {}
    if (data_dir / "animelist_api.json").exists():
        animelist_api_json_data = {
            el.id: el
            for el in iter_api_list(
                str(data_dir / "animelist_api.json"), list_type=ListType.ANIME
            )
        }
    if (data_dir / "mangalist_api.json").exists():
        mangalist_api_json_data = {
            el.id: el
            for el in iter_api_list(
                str(data_dir / "mangalist_api.json"), list_type=ListType.MANGA
            )
        }

    anime_combined_data: Dict[int, AnimeData] = {}

    # combine anime data
    for mal_id, anime_xml in animelist_xml_data.items():

        anime_hist: List[HistoryEntry] = []
        if mal_id in anime_history:
            anime_hist = anime_history[mal_id].entries
            anime_history.pop(mal_id)  # remove from anime history dict

        anime_combined_data[mal_id] = AnimeData(
            username=username,
            XMLData=anime_xml,
            history=anime_hist,
            JSONList=animelist_json_data.pop(mal_id, None),
            APIList=animelist_api_json_data.pop(mal_id, None),
        )

    manga_combined_data: Dict[int, MangaData] = {}
    for mal_id, manga_xml in mangalist_xml_data.items():

        manga_hist: List[HistoryEntry] = []
        if mal_id in manga_history:
            manga_hist = manga_history[mal_id].entries
            manga_history.pop(mal_id)  # remove from manga history dict

        manga_combined_data[mal_id] = MangaData(
            username=username,
            XMLData=manga_xml,
            history=manga_hist,
            JSONList=mangalist_json_data.pop(mal_id, None),
            APIList=mangalist_api_json_data.pop(mal_id, None),
        )

    # while parsing, items were removed when they were merged into
    # the combined data. if anything still exists here, then warn
    if len(animelist_json_data) > 0:
        logger.debug(
            f"animelist_json_data entries left over (likely different parts of export are out of sync or entries on your list were deleted by MAL): {animelist_json_data}"
        )
    if len(mangalist_json_data) > 0:
        logger.debug(
            f"mangalist_json_data entries left over (likely different parts of export are out of sync or entries on your list were deleted by MAL): {mangalist_json_data}"
        )
    # shouldn't be warned for -- if you delete something off your list the
    # local history files still remain -- not sure if the should be deleted
    #
    # if len(anime_history) > 0:
    #    logger.warning(f"anime_history entries left over: {len(anime_history)}")
    # if len(manga_history) > 0:
    #    logger.warning(f"manga_history entries left over: {len(manga_history)}")

    anime_combined = list(anime_combined_data.values())
    manga_combined = list(manga_combined_data.values())

    # e.g. if you had MALEXPORT_COMBINE_FILTER_TAGS="no source,no raws"
    # anything which has those tags would be removed
    # from the results here
    if FILTER_TAGS in os.environ:
        filter_by_tags: List[str] = os.environ[FILTER_TAGS].split(",")

        # if this entry has any tags which are in the filter list
        def filter_func(e: Union[AnimeData, MangaData]) -> bool:
            tags: Set[str] = set(e.tags_list)
            return not any(t in filter_by_tags for t in tags)

        anime_combined = list(filter(filter_func, anime_combined))
        manga_combined = list(filter(filter_func, manga_combined))

    return anime_combined, manga_combined



"""
Selenium singleton and functionality to login to MAL using a chromedriver
"""

import os
import time
import tempfile
import random
import atexit
from pathlib import Path
from functools import lru_cache
from typing import Optional, Dict, Any, Union

import click
from selenium import webdriver as sel
from selenium.webdriver.common.by import By
from selenium.webdriver.firefox.webdriver import WebDriver as Firefox  # type: ignore[import]

from ..paths import LocalDir, _expand_path
from ..log import logger
from ..common import REQUEST_WAIT_TIME

# environment variables to overwrite the location of the chromedriver
# typically this just uses the 'chromedriver' binary,
# as long as thats on your $PATH
HIDDEN_CHROMEDRIVER = bool(int(os.environ.get("MALEXPORT_CHROMEDRIVER_HIDDEN", 0)))
CHROME_LOCATION: Optional[str] = os.environ.get("MALEXPORT_CHROMEDRIVER_LOCATION")

# location for chromedriver to download files to
TEMP_DOWNLOAD_BASE = os.environ.get("MALEXPORT_TEMPDIR", tempfile.gettempdir())

# unique location so we don't use old data
TEMP_DOWNLOAD_DIR = tempfile.mkdtemp(
    dir=_expand_path(Path(TEMP_DOWNLOAD_BASE) / "malexport_driver_downloads")
)

# global so user can edit before a driver is created if they want
CHROME_KWARGS: Dict[str, Any] = {}


BrowserType = str

Browser = Union[sel.Chrome, Firefox]


@lru_cache(maxsize=12)
def webdriver(browser_type: str) -> Union[sel.Chrome, sel.Firefox]:
    bt = browser_type.casefold()
    assert bt in {"chrome", "firefox"}
    if bt == "chrome":
        options = sel.ChromeOptions()
        if HIDDEN_CHROMEDRIVER:
            options.add_argument("headless")  # type: ignore[no-untyped-call]
            options.add_argument("window-size=1920x1080")  # type: ignore[no-untyped-call]
            options.add_argument("disable-gpu")  # type: ignore[no-untyped-call]
        if CHROME_LOCATION is not None:
            CHROME_KWARGS["executable_path"] = CHROME_LOCATION
        else:
            CHROME_KWARGS["executable_path"] = "chromedriver"
        options.add_experimental_option(
            "prefs", {"download.default_directory": str(TEMP_DOWNLOAD_DIR)}
        )
        driver = sel.Chrome(chrome_options=options, **CHROME_KWARGS)
        # quit when python exits to avoid hanging browsers
        atexit.register(lambda: driver.quit())  # type: ignore[no-any-return]
        return driver
    else:
        # mostly added to get around this bug https://github.com/SeleniumHQ/selenium/issues/10799
        # which seems to happen on chromedriver 103 while fetching history
        # hmm -- why is mypy complaining untyped? LSP seems to take me to webdriver definition
        ff = Firefox(  # type: ignore[no-untyped-call]
            service_log_path=os.path.join(tempfile.gettempdir(), "geckodriver.log")
        )
        atexit.register(lambda: ff.quit())
        return ff


LOGIN_PAGE = "https://myanimelist.net/login.php"

LOGIN_ID = "loginUserName"
PASSWORD_ID = "login-password"
LOGIN_BUTTON_CSS = ".inputButton.btn-form-submit[value='Login']"


def driver_login(webdriver: Browser, localdir: LocalDir) -> None:
    """
    Login using the users MAL username and password
    """
    if hasattr(webdriver, "_malexport_logged_in"):
        return
    creds = localdir.load_or_prompt_credentials()
    logger.info(f"Logging into {creds['username']}...")
    time.sleep(1)
    webdriver.get(LOGIN_PAGE)
    time.sleep(1)
    webdriver.find_element(By.ID, LOGIN_ID).send_keys(creds["username"])
    time.sleep(1)
    webdriver.find_element(By.ID, PASSWORD_ID).send_keys(creds["password"])
    time.sleep(1)
    # use script to login incase window is too small to be clickable
    webdriver.execute_script(f"""document.querySelector("{LOGIN_BUTTON_CSS}").click()""")  # type: ignore[no-untyped-call]
    # set marker value on this instance to confirm user has logged in
    if "MALEXPORT_2FA" in os.environ:
        click.confirm(
            "Hit enter once you've logged in with 2FA...",
            prompt_suffix="",
            default=True,
            show_default=False,
        )
    setattr(webdriver, "_malexport_logged_in", True)


# wait a random amount of time to be nice to MAL servers
def wait() -> None:
    time.sleep(REQUEST_WAIT_TIME + random.random() * 4 - 1)



from typing import List

import requests

from ..paths import LocalDir, _expand_file
from ..common import safe_request_json, Json, serialize
from ..log import logger


class FriendDownloader:
    """
    Download friends for a user
    """

    def __init__(self, localdir: LocalDir) -> None:
        self.localdir = localdir
        self.friend_index_path = _expand_file(self.localdir.data_dir / "friends.json")
        self.friend_chunk_size = 100

    def friend_page_url(self, page: int) -> str:
        assert page >= 1
        return f"https://api.jikan.moe/v4/users/{self.localdir.username}/friends?page={page}"

    def download_friend_index(self) -> List[Json]:
        page = 1
        data: List[Json] = []
        while True:
            try:
                new_data = safe_request_json(self.friend_page_url(page))
                assert (
                    "data" in new_data
                ), f"No friends key in Jikan response: {new_data}"
                data.extend(new_data["data"])
                # hit the last page (or there was no data, if we're on the first page)
                if len(new_data["data"]) < self.friend_chunk_size:
                    return data
            except requests.exceptions.RequestException:
                # failed too many times? Assume Jikan failed for some reason,
                # or we've hit a page that doesn't exist (if user had a multiple
                # of friends that was exactly self.friend_chunk_size)
                # we don't want to overwrite if a pagination failed, so return
                # an empty list
                return []
            page += 1

    def update_friend_index(self) -> None:
        """
        Paginate through friends till you hit a chunk of data which has nothing
        """
        friends = self.download_friend_index()
        # in case this was an error, don't overwrite data with nothing
        if len(friends) == 0:
            logger.info(
                f"No friends found for {self.localdir.username} (could've failed to request, or user has no friends), skipping write to file..."
            )
            return
        self.friend_index_path.write_text(serialize(friends))



from .xml import parse_xml
from .mal_list import parse_file as parse_list
from .forum import iter_forum_posts
from .history import iter_user_history
from .friends import iter_friends
from ..list_type import ListType
from .combine import combine
from .api_list import Entry, iter_api_list
from .messages import iter_user_threads



"""
Maintains an authenticated session with MAL
"""

import os
import re
import json
import base64
import webbrowser
from urllib.parse import urlencode, urlparse, parse_qs
from typing import Dict, Any, cast, Iterator

import requests
import click

from ..common import safe_request
from ..log import logger
from ..paths import LocalDir


def _create_pkce_verifier() -> str:
    """
    Creates a random PKCE compliant string. MAL uses a simple method,
    we dont need to create a challenge string
    """
    code_verifier = base64.urlsafe_b64encode(os.urandom(40)).decode("utf-8")
    code_verifier = re.sub("[^a-zA-Z0-9]+", "", code_verifier)
    return code_verifier


MALEXPORT_REDIRECT_URI = os.environ.get("MALEXPORT_REDIRECT_URI", "http://localhost")
MALEXPORT_STATE = "malexport"

LOGIN_BASE = "https://myanimelist.net/v1"


class MalSession:
    """Class to handle requests to MAL"""

    def __init__(self, client_id: str, localdir: LocalDir):
        """
        Creates a new instance of an authenticated MalSession
        """
        self.client_id = client_id
        self.localdir = localdir
        self.session = requests.Session()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(username={self.localdir.username}, client_id={self.client_id})"

    __str__ = __repr__

    def authenticate(self) -> None:
        """
        Update the request Session to include the Bearer token as a header.
        If that doesn't exist, it starts the token auth flow to create one
        """
        refresh_info = self.refresh_info()
        access_token = refresh_info["access_token"]
        logger.debug(f"Access Token: {access_token}")
        self.session.headers.update({"Authorization": f"Bearer {access_token}"})

    def refresh_info(self) -> Dict[str, str]:
        """
        Load or run the OAuth flow to get an access_token for the MAL API
        """
        if self.localdir.refresh_info.exists():
            try:
                return cast(
                    Dict[str, str], json.loads(self.localdir.refresh_info.read_text())
                )
            except json.JSONDecodeError:
                pass
        return self.oauth_flow()

    def oauth_flow(self) -> Dict[str, str]:
        """
        Run the OAuth flow for MALs API
        """
        logger.info("Logging in to MAL...")

        # get a 'code', may need to open a URL to let the user login in the browser
        # and approve their application
        code_verifier = _create_pkce_verifier()
        url = (
            LOGIN_BASE
            + "/oauth2/authorize?"
            + urlencode(
                {
                    "response_type": "code",
                    "client_id": self.client_id,
                    "state": MALEXPORT_STATE,
                    "redirect_uri": MALEXPORT_REDIRECT_URI,
                    "code_challenge": code_verifier,
                    "code_challenge_method": "plain",
                }
            )
        )
        webbrowser.open_new_tab(url)
        click.echo(f"If the URL didn't open automatically, go to\n\n{url}\n")
        redirected_uri = click.prompt(
            "Approve the application, and paste the whole URL it redirects you back here"
        )
        code: str = parse_qs(urlparse(redirected_uri).query)["code"][0]
        logger.debug(f"parsed code: {code}")

        # make the request with HTTP Basic Authentication
        refresh_req = self.session.post(
            LOGIN_BASE + "/oauth2/token",
            headers={
                "Content-Type": "application/x-www-form-urlencoded",
            },
            data={
                "client_id": self.client_id,
                "grant_type": "authorization_code",
                "code": code,
                "redirect_uri": MALEXPORT_REDIRECT_URI,
                "code_verifier": code_verifier,
            },
            auth=(self.client_id, ""),  # leave client secret empty
        )
        refresh_info: Dict[str, str] = refresh_req.json()
        logger.debug(refresh_info)
        if refresh_req.status_code != 200:
            raise RuntimeError(f"Error: {refresh_info}")
        logger.info(f"Saving refresh information to {self.localdir.refresh_info}")
        self.localdir.refresh_info.write_text(refresh_req.text)
        return refresh_info

    def refresh_token(self) -> None:
        old_refresh_info = json.loads(self.localdir.refresh_info.read_text())
        refresh_req = self.session.post(
            LOGIN_BASE + "/oauth2/token",
            headers={
                "Content-Type": "application/x-www-form-urlencoded",
            },
            data={
                "refresh_token": old_refresh_info["refresh_token"],
                "grant_type": "refresh_token",
            },
            auth=(self.client_id, ""),  # leave client secret empty
        )
        if refresh_req.status_code != 200:
            raise RuntimeError(f"Error: {refresh_req.json()}")
        refresh_req.raise_for_status()
        self.localdir.refresh_info.write_text(refresh_req.text)
        logger.debug(refresh_req.json())
        self.authenticate()  # re-authenticate to attach the new access_token

    def refresh_token_if_expired(self, req: requests.Response) -> None:
        """
        Passed to safe_json_request as the error handler, refreshes the token
        if its expired. Currently that lasts for a month
        """
        if req.status_code == 401:
            logger.info("Refreshing token...")
            self.refresh_token()

    def paginate_all_data(self, url: str) -> Iterator[Any]:
        """
        Generic function that works with any MAL url which supports pagination
        Supply limit elsewhere since it may vary per request
        """
        while True:
            resp = self.safe_json_request(url)
            yield resp["data"]
            if "paging" in resp and "next" in resp["paging"]:
                url = resp["paging"]["next"]
            else:
                break

    def safe_request(self, url: str, **kwargs: Any) -> requests.Response:
        """
        A wrapper for the safe_request function -- makes an authenticated request
        to the MAL API
        """
        r: requests.Response = safe_request(
            url,
            session=self.session,
            on_error=self.refresh_token_if_expired,
            wait_time=1,
            **kwargs,
        )
        return r

    def safe_json_request(self, url: str, **kwargs: Any) -> Dict[str, Any]:
        """
        A wrapper for the safe_json_request -- makes an authenticated request
        to the MAL API, waits a bit, parses it to JSON
        """
        return cast(Dict[str, Any], self.safe_request(url, **kwargs).json())



import json
from pathlib import Path
from datetime import datetime, timezone
from typing import NamedTuple, List, Iterator, Optional, Dict, Any

from ..paths import LocalDir


class Message(NamedTuple):
    at: Optional[datetime]
    username: str
    content: str


class Thread(NamedTuple):
    thread_id: int
    subject: str
    messages: List[Message]

    @property
    def url(self) -> str:
        raise NotImplementedError


def iter_user_threads(username: str) -> Iterator[Thread]:
    localdir = LocalDir.from_username(username)
    msg_dir = localdir.data_dir / "messages"
    for file in msg_dir.glob("*.json"):
        yield _parse_thread(file)


def _parse_thread(thread_file: Path) -> Thread:
    thread_id = thread_file.stem
    assert (
        thread_id.isnumeric()
    ), f"Expected thread JSON file where name includes ID, found {thread_file}"
    data = json.loads(thread_file.read_text())
    messages = _parse_messages(data["messages"])
    return Thread(
        thread_id=int(thread_id), messages=list(messages), subject=data["subject"]
    )


def _parse_messages(message_data: List[Dict[str, Any]]) -> Iterator[Message]:
    for msg_data in message_data:
        yield Message(
            at=datetime.fromtimestamp(msg_data["dt"], tz=timezone.utc)
            if msg_data["dt"] is not None
            else None,
            username=msg_data["username"],
            content=msg_data["content"],
        )



"""
Relating to storing credentials and handling location of the data/config directories
"""

import os
import json
from pathlib import Path
from typing import Union, Dict

import yaml
import click

PathIsh = Union[str, Path]

default_local_dir = os.path.join(Path.home(), ".local", "share")
local_directory: str = os.environ.get("XDG_DATA_HOME", default_local_dir)

default_config_dir = os.path.join(Path.home(), ".config")
config_directory: str = os.environ.get("XDG_CONFIG_HOME", default_config_dir)

default_cache_dir = os.path.join(Path.home(), ".cache")
cache_dir: str = os.environ.get("XDG_CACHE_DIR", default_cache_dir)

mal_id_cache_dir = os.environ.get(
    "MAL_ID_CACHE_DIR", os.path.join(cache_dir, "mal-id-cache")
)

default_zip_base = Path(
    os.environ.get(
        "MALEXPORT_ZIP_BACKUPS", os.path.join(local_directory, "malexport_zips")
    )
)

default_zip_base.mkdir(exist_ok=True, parents=True)

default_data_dir = Path(local_directory) / "malexport"
default_conf_dir = Path(config_directory) / "malexport"

if "MALEXPORT_DIR" in os.environ:
    default_data_dir = Path(os.environ["MALEXPORT_DIR"])

if "MALEXPORT_CFG" in os.environ:
    default_conf_dir = Path(os.environ["MALEXPORT_CFG"])

default_data_dir.mkdir(exist_ok=True, parents=True)
default_conf_dir.mkdir(exist_ok=True, parents=True)


def _expand_path(pathish: PathIsh, is_dir: bool = True) -> Path:
    """
    given some path-like input, expand the path
    if is_dir:
        make that directory if it doesnt exist
    else:
        make the parent directory, assuming this is a file
    """
    p: Path
    if isinstance(pathish, str):
        p = Path(pathish)
    else:
        p = pathish
    p = p.expanduser().absolute()
    if is_dir:
        p.mkdir(parents=True, exist_ok=True)
    else:
        p.parent.mkdir(parents=True, exist_ok=True)
    return p


def _expand_file(pathish: PathIsh) -> Path:
    return _expand_path(pathish, is_dir=False)


class LocalDir:
    def __init__(self, application_base: Path, config_base: Path, username: str):
        self.application_base: Path = _expand_path(application_base)
        self.config_base: Path = _expand_path(config_base)
        self.username: str = username

        # to save your personal Client ID from the MAL Dashboard
        self.mal_client_info = self.config_base / "mal_client_id.json"

        # where to save oauth refresh info
        self.refresh_info: Path = (
            _expand_path(self.config_base / "accounts")
            / f"{self.username}_refresh_info.json"
        )

        # To save MAL Username/Password
        self.credential_path: Path = (
            _expand_path(self.config_base / "accounts")
            / f"{self.username}_credentials.yaml"
        )

        # Base directory to store all data
        self.data_dir = _expand_path(self.application_base / self.username)

    def load_or_prompt_mal_client_info(self) -> Dict[str, str]:
        if not self.mal_client_info.exists():
            click.echo(
                "No MAL Client Id/Secret found, create an ID (other/hobbyist) at https://myanimelist.net/apiconfig"
            )
            client_id = click.prompt(text="Client ID")
            with self.mal_client_info.open("w") as f:
                json.dump({"client_id": client_id}, f)
        with self.mal_client_info.open() as f:
            data: Dict[str, str] = json.load(f)
            return data

    def load_or_prompt_credentials(self) -> Dict[str, str]:
        if not self.credential_path.exists():
            click.echo(
                "No credentials found. Enter your username/password for MAL. These are stored locally and used to authenticate a session with MAL",
                err=True,
            )
            username = click.prompt(text="MAL Username")
            password = click.prompt(text="MAL Password")
            with open(self.credential_path, "w") as f:
                yaml.dump({"username": username, "password": password}, f)
                click.echo("Saved to {}".format(self.credential_path))
        with open(self.credential_path) as f:
            data = yaml.load(f, Loader=yaml.FullLoader)
        return dict(
            username=data["username"],
            password=data["password"],
        )

    @staticmethod
    def from_username(
        username: str,
        *,
        data_dir: Path = default_data_dir,
        conf_dir: Path = default_conf_dir,
    ) -> "LocalDir":
        return LocalDir(
            application_base=_expand_path(data_dir),
            config_base=_expand_path(conf_dir),
            username=username,
        )



"""
Requests history (episode/chapter) for an anime/manga entry

This requires authentication (username/password) and selenium, since
this info isn't accessible from the API
"""

import os
import re
import json
import time
import atexit
from itertools import islice
from pathlib import Path
from typing import Tuple, List, Set, Optional, Iterable, Dict, Any
from datetime import datetime

from lxml import html as ht  # type: ignore[import]
from selenium.webdriver.support.ui import WebDriverWait  # type: ignore[import]
from selenium.webdriver.common.by import By  # type: ignore[import]
from selenium.webdriver.support import expected_conditions as EC  # type: ignore[import]

from ..list_type import ListType
from .mal_list import MalList
from .driver import webdriver, driver_login, wait
from .export_downloader import ExportDownloader
from ..log import logger
from ..paths import LocalDir, _expand_path
from ..common import Json, extract_query_value, serialize
from ..parse.xml import parse_xml


HISTORY_URL = "https://myanimelist.net/ajaxtb.php?keepThis=true&detailed{list_type_letter}id={entry_id}&TB_iframe=true&height=420&width=390"


def history_url(list_type: ListType, entry_id: int) -> str:
    """
    Creates the History URL for a particular type/ID
    """
    return HISTORY_URL.format(
        list_type_letter=list_type.value[0].casefold(), entry_id=entry_id
    )


# if we hit these many recently updated entries which
# are the same as the previous then stop requesting
TILL_SAME_LIMIT = int(os.environ.get("MALEXPORT_EPISODE_LIMIT", 5))

EPISODE_COL_REGEX = re.compile(
    r"Ep (\d+), watched on (\d+)\/(\d+)\/(\d+) at (\d+):(\d+)"
)

CHAPTER_COL_REGEX = re.compile(
    r"Chapter (\d+), read on (\d+)\/(\d+)\/(\d+) at (\d+):(\d+)"
)


def _extract_column_data(col_html: str, list_type: ListType) -> Tuple[int, int]:
    """
    Returns the episode/chapter number and the date as epoch time
    """
    chosen_regex = (
        EPISODE_COL_REGEX if list_type == ListType.ANIME else CHAPTER_COL_REGEX
    )
    m = chosen_regex.match(col_html)
    if not m:
        raise RuntimeError(
            f"Could not match episode/chapter/date of out text {col_html}"
        )
    (count, month, day, year, hour, minute) = m.groups()
    # uses local time, so just create a naive datetime and convert
    when = datetime(
        year=int(year),
        month=int(month),
        day=int(day),
        hour=int(hour),
        minute=int(minute),
        second=0,
    )
    return int(count), int(when.timestamp())


class HistoryManager:
    """
    Uses multiple strategies to update history data (episode/chapter watch dates)
    """

    def __init__(
        self,
        list_type: ListType,
        localdir: LocalDir,
        driver_type: str = "chrome",
        till_same_limit: int = TILL_SAME_LIMIT,
        use_merged_file: bool = False,
    ) -> None:
        self.list_type = list_type
        self.localdir = localdir
        # if we request this many items and there is no difference
        # in any of the responses to the previously saved entries,
        # stop requesting
        self.till_same_limit = till_same_limit
        self.use_merged_file = use_merged_file
        self.merged_data: Optional[Dict[str, Any]] = None

        self.history_path: Path
        if self.use_merged_file:
            logger.debug(f"Using merged file for {self.list_type.value}")
            self.history_path = _expand_path(
                self.localdir.data_dir / f"{self.list_type.value}_history.json",
                is_dir=False,
            )
            if self.history_path.exists():
                self.merged_data = json.loads(self.history_path.read_text())
            else:
                self.merged_data = {}
            self._register_atexit()
        else:
            logger.debug("Using individual history files...")
            self.history_path = _expand_path(
                self.localdir.data_dir / "history" / self.list_type.value
            )

        # a list of IDs already requested by this instance, to avoid
        # duplicates across strategies
        self.already_requested: Set[int] = set()

        self.container_id = (
            "chapdetails" if self.list_type == ListType.MANGA else "epdetails"
        )
        self.idprefix = "chaprow" if self.list_type == ListType.MANGA else "eprow"
        self.driver_type = driver_type
        self.driver = webdriver(self.driver_type)

    def authenticate(self) -> None:
        """Logs in to MAL using your MAL username/password"""
        driver_login(webdriver=self.driver, localdir=self.localdir)

    def entry_path(self, entry_id: int) -> Path:
        """Location of the JSON file for this type/ID"""
        if self.use_merged_file:
            return self.history_path
        else:
            return self.history_path / f"{entry_id}.json"

    def has_data(self, entry_id: int) -> bool:
        if self.use_merged_file:
            assert self.merged_data is not None
            return str(entry_id) in self.merged_data
        else:
            return self.entry_path(entry_id).exists()

    def save_data(self, entry_id: int, new_data: Json) -> bool:
        """
        return True if data changed, False otherwise
        """
        if self.use_merged_file:
            logger.debug(f"Saving {entry_id} data to merged JSON...")
            assert self.merged_data is not None
            key = str(entry_id)
            has_new_data = True
            if key in self.merged_data:
                old_data = self.merged_data[str(entry_id)]
                has_new_data = old_data != new_data
            self.merged_data[str(entry_id)] = new_data
            return has_new_data
        else:
            has_new_data = True
            p = self.entry_path(entry_id)
            logger.debug(f"Saving {entry_id} to {p}...")
            if p.exists():
                old_data = json.loads(p.read_text())
                # if these arent the same, the data has changed,
                # we should keep searching for new episode information
                has_new_data = old_data != new_data
            new_data_json = serialize(new_data)
            # this saves even if there is no episode history, so we can compare when updating
            p.write_text(new_data_json)
            return has_new_data

    def _save_merged_file(self) -> None:
        data = serialize(self.merged_data)
        logger.debug(f"Writing merged JSON file: {self.history_path}")
        self.history_path.write_text(data)

    def _register_atexit(self) -> None:
        atexit.register(lambda: self._save_merged_file())

    def _extract_details(self, html_details: str) -> Json:
        """
        Given the HTML div which contains the episode details from the page,
        extract the header (name of the entry) when each episode was watched

        This uses a list instead of an episode -> datetime mapping
        since its possible for you to mark episodes multiple times, e.g. if you're
        rewatching entries
        """
        x = ht.fromstring(html_details)
        # parse the header
        header = x.xpath('.//div[contains(text(), "Details")]')
        assert (
            len(header) == 1
        ), f"Found {len(header)} elements while searching for header, expected 1"
        # split tokens and remove last two (Chapter Details or Episode Details)
        title = header[0].text.strip()
        title = " ".join(title.split(" ")[:-2]).strip()
        data: Dict[str, Any] = {"title": title}
        # parse episodes/chapters; fine even if there are no episode/chapter elements
        episode_elements = x.xpath(f'.//div[starts-with(@id, "{self.idprefix}")]')
        episodes = [
            list(_extract_column_data(ep.text, self.list_type))
            for ep in episode_elements
        ]
        # sort by date, most recent first
        episodes.sort(key=lambda d: d[1], reverse=True)
        data["episodes"] = episodes
        logger.debug(data)
        return data

    def download_recent_user_history(self) -> List[int]:
        """
        Goes to the history page in the authenticated browser
        and grabs IDs for any items which were recently watched/read
        """
        logger.info(f"Downloading recent user {self.list_type.value} history")
        time.sleep(1)
        mal_username = self.localdir.load_or_prompt_credentials()["username"]
        history_url = (
            f"https://myanimelist.net/history/{mal_username}/{self.list_type.value}"
        )
        self.driver.get(history_url)
        wait()
        content_div = self.driver.find_element(By.CSS_SELECTOR, "div#content")
        x = ht.fromstring(content_div.get_attribute("innerHTML"))
        found_ids: List[int] = []
        for el in x.xpath(
            f'.//a[starts-with(@href, "/{self.list_type.value}.php?id=")]'
        ):
            new_id = int(extract_query_value(el.attrib["href"], "id"))
            if new_id not in found_ids:
                found_ids.append(new_id)
        return found_ids

    def download_history_for_entry(self, entry_id: int) -> Json:
        """
        Download the information for a particular type/ID
        """
        time.sleep(1)
        url: str = history_url(self.list_type, entry_id)
        logger.info(f"Requesting history data for {self.list_type.value} {entry_id}")
        self.driver.get(url)
        wait()
        # sanity check to make sure data is present on the page
        WebDriverWait(self.driver, 10).until(  # type: ignore[no-untyped-call]
            EC.text_to_be_present_in_element(  # type: ignore[no-untyped-call]
                (
                    By.ID,
                    self.container_id,
                ),
                "Details",
            )
        )
        details = self.driver.find_element(By.ID, self.container_id)
        assert (
            details is not None
        ), f"Couldn't find details (header) div for {self.list_type.value} {entry_id} {url}"
        new_data = self._extract_details(details.get_attribute("innerHTML"))
        return new_data

    def update_entry_data(self, entry_id: int) -> bool:
        """
        This returns a bool which signifies if data was changed
        If any data was changed/this is new, this returns True
        If data was the same as last time, it returns False
        """
        if entry_id in self.already_requested:
            logger.debug(f"{entry_id} has already been requested, skipping...")
            return False
        new_data = self.download_history_for_entry(entry_id)
        self.already_requested.add(entry_id)
        # if using merged file, save every 10 requests
        if (
            self.use_merged_file is True
            and len(self.already_requested) > 0
            and len(self.already_requested) % 10 == 0
        ):
            self._save_merged_file()
        return self.save_data(entry_id, new_data)

    def update_history(self, count: Optional[int] = None) -> None:
        """
        If data doesn't exist at all for an entry, this requests
        info. Then, after that, this employs two strategies to update history data
            - request items that were recently updated (using &order=5
              on the [anime/manga]list) till we hit
              some limit of unchanged data
            - use the history page (using Jikan) for the user to update
              items that have been watched in the last 3 weeks
        If count is specified, only requests the first 'count' entries
        """
        self.authenticate()

        m = MalList(self.list_type, localdir=self.localdir)
        exp = ExportDownloader(localdir=self.localdir)
        export_file = (
            exp.animelist_path
            if self.list_type == ListType.ANIME
            else exp.mangalist_path
        )
        mal_id: int
        logger.info("Requesting any items which don't exist in history...")
        updated = False
        if m.list_path.exists():
            updated = True
            mlist = m.load_list()
            for entry_data in mlist:
                mal_id = entry_data[f"{self.list_type.value}_id"]
                # If data doesn't exist for an item, request it
                # this doesn't impact the other strategies and will likely run when
                # you first add an item or when this is first run and is caching your entire list
                if not self.has_data(mal_id):
                    self.update_entry_data(mal_id)

            logger.info("Requesting items till we hit some amount of unchanged data...")
            # request some amount till we hit unchanged data
            till = int(self.till_same_limit)
            for entry_data in mlist:
                if till <= 0:
                    break
                logger.info(f"Requesting {till} more entries...")
                mal_id = entry_data[f"{self.list_type.value}_id"]
                if self.update_entry_data(mal_id):
                    logger.debug(
                        f"{self.list_type.value} {mal_id} had new data, resetting..."
                    )
                    till = int(self.till_same_limit)
                else:
                    logger.debug(
                        f"{self.list_type.value} {mal_id} matched old data, decrementing..."
                    )
                    till -= 1

        if export_file.exists():
            updated = True
            # use the XML file if that exists
            for el in parse_xml(str(export_file)).entries:
                if not self.has_data(el.id):
                    self.update_entry_data(el.id)
        if not updated:
            raise RuntimeError(
                f"Neither {m.list_path} (lists) or {export_file} (export) exist, need one to update history"
            )

        recent_history: Iterable[int] = iter(self.download_recent_user_history())
        if count is not None:
            logger.info(f"Requesting {count} first items from user history")
            recent_history = islice(recent_history, count)
        else:
            logger.info("Requesting all items from user history")
        # use selenium to go to users' history and update things watched within the last few weeks
        for mal_id in recent_history:
            self.update_entry_data(mal_id)
        self._save_merged_file()



from typing import NamedTuple, Iterator
from datetime import datetime
import json

from ..paths import LocalDir
from ..log import logger


class Friend(NamedTuple):
    url: str
    username: str
    image_url: str
    friends_since: datetime
    last_online: datetime


def iter_friends(username: str) -> Iterator[Friend]:
    localdir = LocalDir.from_username(username)
    friends_path = localdir.data_dir / "friends.json"
    if not friends_path.exists():
        logger.debug(f"{friends_path} doesn't exist, returning empty iterator")
        return
    for blob in json.loads(friends_path.read_text()):
        user = blob["user"]
        yield Friend(
            url=user["url"],
            username=user["username"],
            image_url=user["images"]["jpg"]["image_url"],
            last_online=datetime.fromisoformat(blob["last_online"]),
            friends_since=datetime.fromisoformat(blob["friends_since"]),
        )



from pathlib import Path
from setuptools import setup, find_packages  # type: ignore[import]

long_description = Path("README.md").read_text()
reqs = Path("requirements.txt").read_text().strip().splitlines()

pkg = "malexport"
setup(
    name=pkg,
    version="0.1.2",
    url="https://github.com/seanbreckenridge/malexport",
    author="Sean Breckenridge",
    author_email="seanbrecke@gmail.com",
    description=("""backs up info from your MAL account"""),
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
    packages=find_packages(),
    install_requires=reqs,
    package_data={pkg: ["py.typed"]},
    zip_safe=False,
    keywords="",
    entry_points={"console_scripts": ["malexport = malexport.__main__:main"]},
    extras_require={
        "testing": [
            "mypy",
        ],
        "recover": ["hpi"],
    },
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
)



import os
import json
import glob
from datetime import datetime
from typing import NamedTuple, Iterator

from ..paths import LocalDir


class Post(NamedTuple):
    title: str
    comment_id: int
    username: str
    url: str
    created_at: datetime
    body: str


def iter_forum_posts(username: str) -> Iterator[Post]:
    localdir = LocalDir.from_username(username)
    mal_username = localdir.load_or_prompt_credentials()["username"]
    forum_dir = localdir.data_dir / "forum"
    for forum_path in glob.glob(os.path.join(forum_dir, "*.json")):
        yield from _extract_posts_by_user(forum_path, mal_username)


def _extract_posts_by_user(forum_path: str, username: str) -> Iterator[Post]:
    forum_id, _ = os.path.splitext(os.path.basename(forum_path))
    if not forum_id.isnumeric():  # not a valid post, probably index.json
        return
    with open(forum_path, "r") as f:
        data = json.load(f)
    for post in data["posts"]:
        if username.casefold() == post["created_by"]["name"].casefold():
            yield Post(
                title=data["title"],
                comment_id=post["id"],
                created_at=datetime.fromisoformat(post["created_at"]),
                username=username,
                url=f"""https://myanimelist.net/forum/?topicid={forum_id}&show=0#msg{post["id"]}""",
                body=post["body"],
            )



"""
Requests messages (DMs)

This requires authentication (username/password) and selenium, since
this info isn't accessible from the API
"""

import os
import json
import time
from pathlib import Path
from typing import List, Optional, Dict, Iterator, Any, Tuple

import dateparser
import more_itertools
from lxml import html as ht, etree  # type: ignore[import]
from selenium.webdriver.common.by import By  # type: ignore[import]

from .driver import webdriver, driver_login, wait
from ..log import logger
from ..paths import LocalDir, _expand_path
from ..common import Json, extract_query_value, serialize


# if we hit these many recently updated entries which
# are the same as the previous then stop requesting
TILL_SAME_LIMIT = int(os.environ.get("MALEXPORT_THREAD_LIMIT", 10))


def dateparse_to_epoch(datestr: str) -> Optional[int]:
    val = dateparser.parse(datestr.strip())
    if val:
        return int(val.timestamp())
    return None


class MessageDownloader:
    def __init__(
        self,
        localdir: LocalDir,
        driver_type: str = "chrome",
        till_same_limit: Optional[int] = None,
    ) -> None:
        self.localdir = localdir
        # if we request this many items and there is no difference
        # in any of the responses to the previously saved entries,
        # stop requesting
        self.till_same_limit: int = int(till_same_limit or TILL_SAME_LIMIT)
        self.message_base_path: Path = _expand_path(self.localdir.data_dir / "messages")
        self.msg_to_thread: Dict[int, int] = {}

        self.driver_type = driver_type
        self.driver = webdriver(self.driver_type)

    def authenticate(self) -> None:
        """Logs in to MAL using your MAL username/password"""
        driver_login(webdriver=self.driver, localdir=self.localdir)

    def entry_path(self, thread_id: int) -> Path:
        """Location of the JSON file for this thread"""
        return self.message_base_path / f"{thread_id}.json"

    def _fix_subject(self, subject: str) -> str:
        s = subject.strip()
        if s.startswith("re:"):
            s = s[len("re:") :]
        return s.strip()

    def _extract_details(self, html_details: str) -> Json:
        """
        Given the HTML div which contains the container for messages,
        parse it into JSON

        """
        hx = ht.fromstring(html_details)
        table = hx.cssselect("table.pmessage-message-history")[0]
        subject = hx.cssselect(".dialog-text .mb4")[-1]
        data: Dict[str, Any] = {
            "messages": [],
            "subject": self._fix_subject(subject.text_content()),
        }
        for row in table.cssselect("tr"):
            date_td = row.cssselect("td.date")[0]
            username_td = row.cssselect("td.name")[0]
            content_td = row.cssselect("td.subject")[0]
            data["messages"].append(
                {
                    "username": username_td.text_content().strip(),
                    "dt": dateparse_to_epoch(date_td.text_content()),
                    "content": etree.tostring(content_td).decode("utf-8").strip(),
                }
            )
        logger.debug(data)
        return data

    def message_ids_for_page(self, page: int = 1, sent: bool = False) -> List[int]:
        """
        Goes to the message ids for a particular page of your messages
        """
        logger.info(
            f"Downloading page {page} of your {'sent ' if sent else ''}messages"
        )
        time.sleep(1)
        offset = (page - 1) * 20
        message_url = f"https://myanimelist.net/mymessages.php?go={'sent' if sent else ''}&show={offset}"
        self.driver.get(message_url)
        wait()
        # extract id=349234 from each message URL
        return [
            int(str(extract_query_value(a.get_attribute("href"), "id")))
            for a in self.driver.find_elements(By.CSS_SELECTOR, "a.subject-link")
        ]

    def iter_message_ids(self, start_page: int = 1) -> Iterator[Tuple[bool, int]]:
        """
        lazily retrieves new messages IDs for your user
        swaps between received messages and sent messages, so that
        this roughly grabs messages from the same time frame at the same time
        """
        page = int(start_page)
        while True:
            message_ids: List[int] = self.message_ids_for_page(page)
            sent_ids: List[int] = self.message_ids_for_page(page, sent=True)
            tagged_msg_id = [(False, mid) for mid in message_ids]
            tagged_sent_ids = [(True, mid) for mid in sent_ids]
            for sent, msg_id in more_itertools.interleave_longest(
                tagged_msg_id, tagged_sent_ids
            ):
                yield sent, msg_id
            if len(message_ids) == 0 and len(sent_ids) == 0:
                return
            page += 1

    def _resolve_message_to_thread_id(self, message_id: int, sent: bool) -> int:
        """
        goes to a message ID page and clicks the 'view message history' button
        returns the thread ID this corresponds to
        """
        time.sleep(1)
        url: str = f"https://myanimelist.net/mymessages.php?go=read&id={message_id}{'&f=1' if sent else ''}"
        logger.debug(f"Resolving message ID {message_id} to thread...")
        logger.debug(f"Navigating to '{url}'")
        self.driver.get(url)
        wait()
        thread_link = self.driver.find_element(
            By.PARTIAL_LINK_TEXT, "View Message History"
        )
        thread_url = thread_link.get_attribute("href")
        self.driver.get(thread_url)
        wait()
        return int(str(extract_query_value(thread_url, "threadid")))

    def update_thread_data(self, thread_id: int) -> bool:
        """
        This returns a bool which signifies if data was changed
        If any data was changed/this is new, this returns True
        If data was the same as last time, it returns False
        """
        if thread_id in self.msg_to_thread.values():
            logger.debug(f"thread {thread_id} has already been requested, skipping...")
            return False
        p = self.entry_path(thread_id)
        # at this point, we're already on the thread page
        new_data = self._extract_details(
            self.driver.find_element(By.ID, "content").get_attribute("innerHTML")
        )
        # assume this is new data
        has_new_data = True
        if p.exists():
            old_data = json.loads(p.read_text())
            # if these arent the same, the data has changed,
            # we should keep searching for more threads
            has_new_data = old_data != new_data
        new_data_json = serialize(new_data)
        p.write_text(new_data_json)
        return has_new_data

    def update_messages(
        self, start_page: int = 1, thread_count: Optional[int] = None
    ) -> None:
        self.authenticate()

        # if user supplied with CLI flag use that, else use envvar/default
        till_base = (
            int(self.till_same_limit) if thread_count is None else int(thread_count)
        )
        till = int(till_base)

        for sent, message_id in self.iter_message_ids(start_page=start_page):
            if till <= 0:
                break
            # resolve message ID to thread, which is what we download
            thread_id: int = self._resolve_message_to_thread_id(message_id, sent=sent)
            logger.info(f"msg {message_id} -> thread {thread_id}")
            # keep track of if this is a new thread, so we can decrement the 'till same' counter
            new_thread_id: bool = thread_id not in self.msg_to_thread.values()
            if self.update_thread_data(thread_id):
                logger.debug(
                    f"msg id {message_id}, thread {thread_id} had new data, resetting..."
                )
                till = int(till_base)
            else:
                if new_thread_id:
                    logger.debug(
                        f"msg id {message_id} thread {thread_id} matched old data, decrementing..."
                    )
                    till -= 1
            logger.info(f"requesting {till} more threads...")
            # save thread id so subsequent iterations dont repeat
            self.msg_to_thread[message_id] = thread_id
